#include "precompile.h"

#include <f8/timer.h>

#include "matchteam.h"
#include "matchmgr.h"
#include "GGListener.h"
#include "roommgr.h"
#include "mt/Param.h"
#include "mt/Robot.h"

#include "f8/utils.h"

#include "cs_proto.pb.h"

RawTeamMember::RawTeamMember()
{
    msg = std::make_shared<cs::CMJoin>();
    battle_context = std::make_shared<BattleDataContext>();
}

void RawTeamMember::FillMFMatchTeamMember(cs::MFMatchTeamMember* p)
{
    p->set_account_id(msg->account_id());
    p->set_name(msg->name());
    p->set_avatar_url(msg->avatar_url());
    #if 1
    {
        p->set_hero_id(msg->hero_id());
        *p->mutable_weapons() = msg->weapons();
    }
    #else
    if (choose_hero_times <= 0) {
        p->set_hero_id(0);
    } else {
        p->set_hero_id(msg->hero_id());
        *p->mutable_weapons() = msg->weapons();
    }
    #endif
    p->set_is_leader(is_leader);
    p->set_state(state);
    p->set_head_frame(msg->head_frame());
}

void RawTeamMember::InitRobot()
{
    static long long robot_idx = 1000;

    msg->set_account_id(a8::Format("6000_2006_%d", {++robot_idx}));
    {
        std::set<int> refreshed_robot_set;
        robot_meta = mt::Robot::RandRobot(refreshed_robot_set);
        if (robot_meta) {
            msg->set_avatar_url(a8::Format("%s", {50001 + (robot_meta->id() % 3)}));
            msg->set_hero_id(30100 + 100 * (robot_meta->id() % 3));
            msg->set_name(robot_meta->name());
        } else {
            A8_ABORT();
        }
    }
}

MatchTeam::~MatchTeam()
{
    timer_attacher.ClearTimerList();
#ifdef DEBUG
    a8::XPrintf("~MatchTeam %s\n", {GetTeamUUid()});
#endif
}

void MatchTeam::Init(f8::MsgHdr& hdr, const cs::CMJoin& msg)
{
    master_team_ = this;
    create_tick_ = a8::XGetTickCount();
    f8::Timer::Instance()->SetIntervalEx
        (1000,
         [this] (int event, const a8::Args* args)
         {
             if (a8::TIMER_EXEC_EVENT == event) {
                 Update();
             }
         },
         &timer_attacher);
    phase_= kMatchCombining;
    phase_start_tick_ = a8::XGetTickCount();
    countdown_ = mt::Param::s().match_team_time;
    AddRawMember(hdr, msg);
}

void MatchTeam::_CMMatchCancel(f8::MsgHdr& hdr, const cs::CMMatchCancel& msg)
{
}

void MatchTeam::_CMMatchChoose(f8::MsgHdr& hdr, const cs::CMMatchChoose& msg)
{
#ifdef DEBUG
    a8::XPrintf("cmmatchchoose %s phase_:%d %s\n", {hdr.socket_handle, phase_, GetTeamUUid()});
    {
        for (auto& pair : raw_member_hash_) {
            a8::XPrintf("cmmatchchoose member %d:%d:%s %s:%d\n",
                        {
                         hdr.socket_handle,
                         phase_,
                         GetTeamUUid(),
                         pair.second->msg->account_id(),
                         pair.second->socket_handle
                        });
        }
    }
#endif
    auto member = GetMemberBySocket(hdr.socket_handle);
    if (member && master_team_->phase_ == kMatchChoose) {
        member->msg->set_hero_id(msg.hero_id());
        member->msg->set_hero_uniid(msg.hero_uniid());
        *member->msg->mutable_weapons() = msg.weapons();
        ++member->choose_hero_times;
#ifdef DEBUG
        a8::XPrintf("matchchoose %s\n", {member->msg->account_id()});
#endif
    }
}

void MatchTeam::_CMMatchStartGame(f8::MsgHdr& hdr, const cs::CMMatchStartGame& msg)
{
    if (master_team_->phase_ == kMatchChoose) {
        auto member = GetMemberBySocket(hdr.socket_handle);
        if (member) {
            member->state = kMatchPrepare;
#ifdef DEBUG
        a8::XPrintf("matchgamestart %s\n", {member->msg->account_id()});
#endif
        }
    }
}

void MatchTeam::_CMMatchCancelStartGame(f8::MsgHdr& hdr, const cs::CMMatchCancelStartGame& msg)
{
    if (master_team_->phase_ == kMatchChoose) {
        auto member = GetMemberBySocket(hdr.socket_handle);
        if (member) {
            member->state = kMatchReadying;
#ifdef DEBUG
        a8::XPrintf("matchcancelgamestart %s\n", {member->msg->account_id()});
#endif
        }
    }
}

void MatchTeam::_CMMatchSendMsg(f8::MsgHdr& hdr, const cs::CMMatchSendMsg& msg)
{
    auto sender = GetMemberBySocket(hdr.socket_handle);
    if (sender) {
        for (auto member : master_team_->curr_member_hash_) {
            if (member->socket_handle != 0) {
                bool found = false;
                for (auto& target : msg.target_list()) {
                    if (target == member->msg->account_id()) {
                        found = true;
                        break;
                    }
                }
                if (found) {
                    cs::SMMatchMemberMsgNotify notifymsg;
                    notifymsg.set_sender(sender->msg->account_id());
                    notifymsg.set_content(msg.content());
                    GGListener::Instance()->SendToClient(member->socket_handle, 0, notifymsg);
                }
            }
        }
    }
}

void MatchTeam::_CMMatchBroadcastMsg(f8::MsgHdr& hdr, const cs::CMMatchBroadcastMsg& msg)
{
    auto sender = GetMemberBySocket(hdr.socket_handle);
    if (sender) {
        for (auto member : master_team_->curr_member_hash_) {
            if (member->socket_handle != 0) {
                if (!msg.exclude_self() ||
                    (msg.exclude_self() && sender->msg->account_id() != member->msg->account_id())) {
                    cs::SMMatchMemberMsgNotify notifymsg;
                    notifymsg.set_sender(sender->msg->account_id());
                    notifymsg.set_content(msg.content());
                    GGListener::Instance()->SendToClient(member->socket_handle, 0, notifymsg);
                }
            }
        }
    }
}

void MatchTeam::AddRawMember(f8::MsgHdr& hdr, const cs::CMJoin& msg)
{
    std::shared_ptr<RawTeamMember> member = std::make_shared<RawTeamMember>();
    member->team = this;
    member->add_tick = a8::XGetTickCount();
    member->socket_handle = hdr.socket_handle;
    *member->msg = msg;
    raw_member_hash_[msg.account_id()] = member;
    curr_member_hash_.push_back(member);
    if (!first_member_) {
        first_member_ = member;
    }
    if (master_team_ != this) {
        master_team_->curr_member_hash_.push_back(member);
    }
    #ifdef DEBUG
    a8::XPrintf("AddRawMember %s %s %d %d %d\n",
                {
                    GetTeamUUid(),
                        msg.account_id(),
                        curr_member_hash_.size(),
                        master_team_->curr_member_hash_.size(),
                        master_team_ == this ? 1 : 0
        });
    #endif
}

bool MatchTeam::IsRawMember(const std::string &account_id)
{
    auto itr = raw_member_hash_.find(account_id);
    return itr != raw_member_hash_.end();
}

bool MatchTeam::IsValidMember(const cs::CMJoin& msg)
{
    if (first_member_->msg->team_uuid() != msg.team_uuid()) {
        return false;
    }
    if (first_member_->msg->room_mode() != msg.room_mode()) {
        return false;
    }
    if (first_member_->msg->pve_instance_id() != msg.pve_instance_id()) {
        return false;
    }
    if (first_member_->msg->team_slot_num() != msg.team_slot_num()) {
        return false;
    }
    if (first_member_->msg->team_members().size() != msg.team_members().size()) {
        return false;
    }
    for (int i = 0; i < msg.team_members().size(); ++i) {
        auto& a = msg.team_members(i);
        auto& b = first_member_->msg->team_members(i);
        if (a.account_id() != b.account_id()) {
            return false;
        }
    }
    return true;
}

void MatchTeam::Update()
{
    if (phase_ == kMatchWaitStart) {
        return;
    }
    if (IsMasterTeam()) {
        UpdateMaster();
    } else {
        UpdateSlave();
    }
    if (phase_ == kMatchStartGame && IsMasterTeam()) {
        StartGame();
    }
}

void MatchTeam::SyncMatchInfo()
{
    cs::SMUpdateMatchInfo notifymsg;
    for (auto member : curr_member_hash_) {
        if (phase_ > kMatchCombining || member->team == this) {
            member->FillMFMatchTeamMember(notifymsg.mutable_info()->add_members());
        }
    }
    notifymsg.mutable_info()->set_phase(phase_);
    notifymsg.mutable_info()->set_countdown(phase_left_time_);
    notifymsg.mutable_info()->set_predict_time(mt::Param::s().match_team_time);
    for (auto& member : curr_member_hash_) {
        if (member->socket_handle != 0) {
            GGListener::Instance()->SendToClient(member->socket_handle, 0, notifymsg);
#ifdef DEBUG
            {
                a8::XPrintf("SyncMatchInfo sokcet_handle:%d account_id:%s\n",
                            {
                                member->socket_handle,
                                    member->msg->account_id()
                            });
            }
#endif
        }
    }
#ifdef DEBUG
    {
        std::string data = f8::PbToJson(&notifymsg);
        a8::XPrintf("SMUpdateMatchInfo %s\n", {data});
    }
#endif
}

void MatchTeam::TryCombineTeam()
{
    if (HaveEmptySlot()) {
        std::shared_ptr<MatchTeam> matched_team;
        MatchMgr::Instance()->TraverseTeam
            (
             [this, &matched_team] (std::shared_ptr<MatchTeam>& team, bool& stop)
             {
                 if (team.get() == this) {
                     return;
                 }
                 if (CanCombine(team.get())) {
                     matched_team = team;
                     stop = true;
                 }
             });
        if (matched_team) {
            Combine(matched_team.get());
        }
    }
}

void MatchTeam::UpdateMaster()
{
    phase_left_time_ = GetPhaseLeftTime();
    switch (phase_) {
    case kMatchCombining:
        {
            if (IsShuaRobotTime()) {
                if (HaveEmptySlot()) {
                    ShuaRobot();
                }
            } else {
                if (HaveEmptySlot()) {
                    TryCombineTeam();
                }
            }
            if (phase_left_time_ <= 0 || !HaveEmptySlot()) {
                ChooseLeader();
                phase_ = kMatchChoose;
                phase_start_tick_ = a8::XGetTickCount();
                countdown_ = mt::Param::s().match_choose_time;
            }
        }
        break;
    case kMatchChoose:
        {
            UpdateTeamState();
            CheckChoose();
            CheckPrepare();
            if ((phase_left_time_ <= mt::Param::s().match_lock_time) ||
                IsAllPrepare()) {
                AutoChoose(true);
                for (auto& member : curr_member_hash_) {
                    member->state = kMatchPrepare;
                }
                phase_ = kMatchLock;
                phase_start_tick_ = a8::XGetTickCount();
                countdown_ = mt::Param::s().match_lock_time;
            }
        }
        break;
    case kMatchLock:
        {
            if (phase_left_time_ <= 0) {
                phase_ = kMatchStartGame;
                phase_start_tick_ = a8::XGetTickCount();
            }
        }
        break;
    case kMatchStartGame:
        {
        }
        break;
    default:
        {
        }
        break;
    }
    master_team_->SyncMatchInfo();
}

void MatchTeam::UpdateSlave()
{
    switch (phase_) {
    case kMatchCombining:
        {
        }
        break;
    case kMatchChoose:
        {

        }
        break;
    case kMatchLock:
        {

        }
        break;
    default:
        {
        }
        break;
    }
}

bool MatchTeam::HasSameCurrMember(MatchTeam* b)
{
    bool has = false;
    for (auto& a_member : curr_member_hash_) {
        for (auto& b_member : b->curr_member_hash_) {
            if (a_member->msg->account_id() == b_member->msg->account_id()) {
                has = true;
                break;
            }
        }
    }
    return has;
}

bool MatchTeam::ProtoIsCompatible(MatchTeam* b)
{
    bool is_compatible = true;
    for (auto& a_member : curr_member_hash_) {
        for (auto& b_member : b->curr_member_hash_) {
            if (a_member->msg->proto_version() != 0 &&
                b_member->msg->proto_version() != 0 &&
                #ifdef DEBUG
                (
                 (a_member->msg->proto_version() < 2022032201 && b_member->msg->proto_version() >= 2022032201) ||
                 (a_member->msg->proto_version() >= 2022032201 && b_member->msg->proto_version() < 2022032201)
                 )
                #else
                false
                #endif
                ) {
                is_compatible = false;
                break;
            }
        }
    }
    return is_compatible;
}

bool MatchTeam::CanCombine(MatchTeam* b)
{
    if (this == b) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 1\n", {});
#endif
        return false;
    }
    //已合并
    if (master_team_ != this){
#ifdef DEBUG
        a8::XPrintf("CanCombine 2\n", {});
#endif
        return false;
    }
    //已合并
    if (b->master_team_ != b) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 3\n", {});
#endif
        return false;
    }
    if (phase_ != kMatchCombining) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 4\n", {});
#endif
        return false;
    }
    if (b->phase_ != kMatchCombining) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 5\n", {});
#endif
        return false;
    }
    if (GetOwner()->msg->room_mode() != b->GetOwner()->msg->room_mode() ||
        GetOwner()->msg->pve_instance_id() != b->GetOwner()->msg->pve_instance_id()) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 31\n", {});
#endif
        return false;
    }
    if (GetSlotNum() != b->GetSlotNum()) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 32\n", {});
#endif
        return false;
    }
    if (IsShuaRobotTime()) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 6\n", {});
#endif
        return false;
    }
    if (b->IsShuaRobotTime()) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 7\n", {});
#endif
        return false;
    }
    if (!b->combined_team_hash_.empty()) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 8\n", {});
#endif
        return false;
    }
    if (!HaveEmptySlot()) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 32\n", {});
#endif
        return false;
    }
    if (!b->HaveEmptySlot()) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 33\n", {});
#endif
        return false;
    }
    if (GetEmptySlotNum() + b->GetInitMemberNum() > GetSlotNum()) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 9\n", {});
#endif
        return false;
    }
    if (combined_team_hash_.find(b->GetTeamUUid()) != combined_team_hash_.end()) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 10\n", {});
#endif
        return false;
    }
    if (b->combined_team_hash_.find(GetTeamUUid()) != b->combined_team_hash_.end()) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 10\n", {});
#endif
        return false;
    }
    if (GetValidSocketNum() <= 0) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 21\n", {});
#endif
        return false;
    }
    if (b->GetValidSocketNum() <= 0) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 22\n", {});
#endif
        return false;
    }
    if (HasSameCurrMember(b)) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 11\n", {});
#endif
        return false;
    }
    if (!ProtoIsCompatible(b)) {
#ifdef DEBUG
        a8::XPrintf("CanCombine 12\n", {});
#endif
        return false;
    }
    return true;
}

void MatchTeam::Combine(MatchTeam* b)
{
#ifdef DEBUG
    a8::XPrintf("Combine %s->%s %d:%d",
                {
                    b->GetTeamUUid(),
                        GetTeamUUid(),
                        b->curr_member_hash_.size(),
                        curr_member_hash_.size()
                        });
#endif
    combined_team_hash_[b->GetTeamUUid()] = b;
    b->master_team_ = this;
    for (auto& member : b->curr_member_hash_) {
        curr_member_hash_.push_back(member);
    }
}

std::string MatchTeam::GetTeamUUid()
{
    return first_member_->msg->team_uuid();
}

int MatchTeam::GetMapId()
{
    return first_member_->msg->mapid();
}

bool MatchTeam::IsShuaRobotTime()
{
    return phase_ == kMatchCombining &&
        phase_left_time_ <= mt::Param::s().match_robot_time;
}

int MatchTeam::GetPhaseLeftTime()
{
    int passtime = (a8::XGetTickCount() - phase_start_tick_) / 1000;
    return std::max(0, countdown_ - passtime);
}

std::shared_ptr<RawTeamMember> MatchTeam::GetMemberBySocket(int socket_handle)
{
    for (auto& pair : raw_member_hash_) {
        if (pair.second->socket_handle == socket_handle) {
            return pair.second;
        }
    }
    return std::shared_ptr<RawTeamMember>();
}

void MatchTeam::ShuaRobot()
{
    for (int i = curr_member_hash_.size(); i < GetSlotNum(); ++i) {
        std::shared_ptr<RawTeamMember> member = std::make_shared<RawTeamMember>();
        member->team = this;
        member->add_tick = a8::XGetTickCount();
        member->is_robot = true;
        member->InitRobot();
        curr_member_hash_.push_back(member);
    }
}

void MatchTeam::StartGame()
{
    phase_ = kMatchWaitStart;
    if (GetValidSocketNum() > 0) {
        std::vector<std::shared_ptr<cs::CMJoin>> join_msgs;
        for (auto& member : GetCurrMembers()) {
            if (!member->is_robot) {
                join_msgs.push_back(member->msg);
            }
        }
        std::string team_uuid = GetTeamUUid();
        auto cb =
            [team_uuid] (std::vector<std::shared_ptr<BattleDataContext>>& results)
            {
                auto team = MatchMgr::Instance()->GetTeam(team_uuid);
                if (team) {
                    for (auto context : results) {
                        bool found = false;
                        for (auto& member : team->GetCurrMembers()) {
                            if (member->msg->account_id() == context->join_msg->account_id()){
                                member->battle_uuid = context->battle_uuid;
                                member->is_valid_battle = context->is_valid_battle;
                                member->payload = context->payload;
                                member->battle_context = context;
                                found = true;
                                break;
                            }
                        }
                        if (!found) {
                            abort();
                        }
                    }
                    RoomMgr::Instance()->JoinTeam(team.get());
                    team->RemoveTeam();
                }
            };
        RoomMgr::Instance()->SendGetBattleData(1, join_msgs, cb);
    } else {
        RemoveTeam();
    }
}

void MatchTeam::ChooseLeader()
{
    std::shared_ptr<RawTeamMember> leader;
    for (auto& member : curr_member_hash_) {
        if (!member->is_robot) {
            if (member->socket_handle) {
                if (!leader) {
                    leader = member;
                    leader->is_leader = true;
                    continue;
                }
            }
        }
        member->is_leader = false;
    }
}

void MatchTeam::AutoChoose(bool force)
{
    for (auto& member : curr_member_hash_) {
        if (member->choose_hero_times <= 0) {
            if (member->is_robot){
                ++member->choose_hero_times;
                break;
            } else if (force) {
                ++member->choose_hero_times;
                break;
            }
        }
    }
}

bool MatchTeam::IsAllPrepare()
{
    int num = 0;
    for (auto& member : curr_member_hash_) {
        if (member->state == kMatchPrepare) {
            ++num;
        }
    }
    return num >= GetSlotNum();
}

void MatchTeam::UpdateTeamState()
{
    {
        bool all_choose = true;
        for (auto& member : curr_member_hash_) {
            if (!member->is_robot && member->choose_hero_times <= 0 && member->socket_handle != 0) {
                all_choose = false;
                break;
            }
        }
        if (all_choose) {
            ++all_player_choose_times_;
        }
    }
    {
        bool all_prepare = true;
        for (auto& member : curr_member_hash_) {
            if (!member->is_robot && member->state != kMatchPrepare && member->socket_handle != 0) {
                all_prepare = false;
                break;
            }
        }
        if (all_prepare) {
            ++all_player_prepare_times_;
        }
    }
}

void MatchTeam::CheckChoose()
{
    if (phase_ != kMatchChoose) {
        return;
    }

    if (phase_left_time_ <= 10 || all_player_choose_times_ > 0) {
        if (last_auto_choose_tick_ <= 0) {
            AutoChoose();
            last_auto_choose_tick_ = a8::XGetTickCount();
        } else {
            if (a8::XGetTickCount() - last_auto_choose_tick_ > (last_auto_choose_tick_ % 2000)) {
                AutoChoose();
                last_auto_choose_tick_ = a8::XGetTickCount();
            }
        }
    }
}

void MatchTeam::CheckPrepare()
{
    if (phase_ != kMatchChoose) {
        return;
    }

    if (phase_left_time_ <= 10 || all_player_prepare_times_ > 0) {
        if (last_auto_prepare_tick_ <= 0) {
            AutoPrepare();
            last_auto_prepare_tick_ = a8::XGetTickCount();
        } else {
            if (a8::XGetTickCount() - last_auto_prepare_tick_ > (last_auto_prepare_tick_ % 2000)) {
                AutoPrepare();
                last_auto_prepare_tick_ = a8::XGetTickCount();
            }
        }
    }
}

void MatchTeam::AutoPrepare()
{
    for (auto& member : curr_member_hash_) {
        if (member->choose_hero_times > 0 &&
            member->state != kMatchPrepare &&
            member->socket_handle == 0) {
            member->state = kMatchPrepare;
            break;
        }
    }
}

int MatchTeam::GetValidSocketNum()
{
    int valid_socket_num = 0;
    for (auto& member : curr_member_hash_) {
        if (member->socket_handle != 0) {
            if (MatchMgr::Instance()->GetMatchInfo(member->socket_handle)) {
                ++valid_socket_num;
            }
        }
    }
    return valid_socket_num;
}

void MatchTeam::RemoveTeam()
{
    for (auto& member : curr_member_hash_) {
        if (member->socket_handle != 0) {
            MatchMgr::Instance()->RemoveSocket(member->socket_handle);
        }
    }
    for (auto& pair : combined_team_hash_) {
        MatchMgr::Instance()->RemoveTeam(pair.first);
    }
    MatchMgr::Instance()->RemoveTeam(GetTeamUUid());
}

int MatchTeam::GetInitMemberNum()
{
    return GetOwner()->msg->team_members().size();
}

int MatchTeam::GetSlotNum()
{
    return GetOwner()->msg->team_slot_num();
}

bool MatchTeam::HaveEmptySlot()
{
    return GetEmptySlotNum() > 0;
}

int MatchTeam::GetEmptySlotNum()
{
    int alloced_slot_num = GetInitMemberNum();
    for (auto& pair : combined_team_hash_) {
        alloced_slot_num += pair.second->GetInitMemberNum();
    }
    int empty_slot_num = std::max(0, GetSlotNum() - alloced_slot_num);
    return empty_slot_num;
}
