#include "precompile.h"

#include <random>

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>

#include <a8/mutable_xobject.h>
#include <a8/collision.h>

#include <f8/udplog.h>
#include <f8/utils.h>
#include <f8/timer.h>
#include <f8/btmgr.h>

#include "playermgr.h"
#include "player.h"
#include "room.h"
#include "android.h"
#include "bullet.h"
#include "obstacle.h"
#include "roomobstacle.h"
#include "loot.h"
#include "car.h"
#include "hero.h"
#include "roommgr.h"
#include "app.h"
#include "gamelog.h"
#include "typeconvert.h"
#include "entityfactory.h"
#include "perfmonitor.h"
#include "mapinstance.h"
#include "mapmgr.h"
#include "incubator.h"
#include "team.h"
#include "matchteam.h"
#include "frameeventdata.h"
#include "pbutils.h"
#include "movement.h"
#include "bornpoint.h"
#include "airdrop.h"
#include "airraid.h"
#include "sandtable.h"
#include "batchsync.h"
#include "trigger.h"
#include "custom_battle.h"
#include "custom_team.h"
#include "custom_member.h"
#include "room_agent.h"

#include "mt/Param.h"
#include "mt/Hero.h"
#include "mt/Equip.h"
#include "mt/Buff.h"
#include "mt/Skill.h"
#include "mt/SkillNumber.h"
#include "mt/GunTalentGrow.h"
#include "mt/SafeArea.h"
#include "mt/MapThing.h"
#include "mt/Text.h"
#include "mt/Map.h"
#include "mt/PveGemini.h"
#include "mt/PveGeminiMode.h"
#include "mt/RankReward.h"
#include "mt/RankPoint.h"
#include "mt/KillReward.h"
#include "mt/KillPoint.h"
#include "mt/Drop.h"
#include "mt/AirLine.h"
#include "mt/Robot.h"
#include "mt/SafeAreaPos.h"
#include "mt/MapThingGroup.h"

const int SHUA_RANGE = 580;

static long long RoomXGetTickCount(void* context)
{
    Room* room = (Room*)context;
    return room->GetFrameNo();
}

Room::~Room()
{

}

void Room::InitData(RoomInitInfo& init_info)
{
    room_idx_ = init_info.room_idx;
    room_mode_ = init_info.room_mode;
    room_uuid_ = init_info.room_uuid;
    room_type_ = init_info.room_type;
    creator_game_times_ = init_info.creator_game_times;
    creator_register_time_ = init_info.creator_register_time;
    creator_proto_version_ = init_info.creator_proto_version;
    creator_channel_ = init_info.creator_channel;
    force_entry_newbie_room_ = init_info.force_entry_newbie_room;
    pve_human_num = init_info.pve_human_num;
    is_newbie_room_ = init_info.is_newbie_room;

    map_tpl_name_ = init_info.map_tpl_name;
    grid_service = init_info.grid_service;
    map_service = init_info.map_service;
    map_instance = init_info.map_instance;
    map_meta_ = init_info.map_meta;
    poly_ext_datas_ = map_instance->GetPolyExtDatas();
    custom_battle_ = init_info.custom_battle;
    rank_match_conf_ = mt::Param::GetRankModeConfByRoomType(room_type_);
}

void Room::Init()
{
    xtimer.Init(RoomXGetTickCount, this, 100, 100);
    xtimer_attacher_.SetOwner(&xtimer);

    frame_event.room = this;
    frame_event_data = std::make_shared<FrameEventData>();

    air_drop_ = std::make_shared<AirDrop>(this);
    air_raid_ = std::make_shared<AirRaid>(this);
    batch_sync_ = std::make_shared<BatchSync>(this);
    room_agent_ = behaviac::Agent::Create<RoomAgent>();
    room_agent_->SetRoom(this);
    f8::BtMgr::Instance()->BtLoad(room_agent_, "npc/room");
    f8::BtMgr::Instance()->BtSetCurrent(room_agent_, "npc/room");

    CreateSpawnPoints();
    CreateWorldObjects();
    ShuaAndroid();
    incubator_ = std::make_shared<Incubator>();
    incubator_->room = this;
    incubator_->Init();
    sand_table_ = std::make_shared<SandTable>(this);
#ifdef DEBUG
    InitDebugInfo();
#endif
    if (IsMobaModeRoom()) {
        moba_over_timer = xtimer.SetTimeoutWpEx
            (SERVER_FRAME_RATE * mt::Param::s().moba_room_time,
             [this] (int event, const a8::Args* args)
             {
                 if (a8::TIMER_EXEC_EVENT == event) {
                     if (!IsGameOver()) {
                         game_over_ = true;
                         game_over_frameno_ = GetFrameNo();
                         OnGameOver();
                     }
                 }
             },
             &xtimer_attacher_);
    }
}

void Room::UnInit()
{
    destorying_ = true;
    incubator_->UnInit();
#ifdef DEBUG
    UnInitDebugInfo();
#endif
    incubator_ = nullptr;
    timer_attacher.ClearTimerList();
    xtimer_attacher_.ClearTimerList();
    grid_service->ClearRoomData(this);
    for (auto& pair : accountid_hash_) {
        PlayerMgr::Instance()->RemovePlayerBySocket(pair.second->socket_handle);
    }
    for (auto& pair : uniid_hash_) {
        delete pair.second;
    }
    uniid_hash_.clear();
    for (auto& pair : later_add_hash_) {
        delete pair.second;
    }
    later_add_hash_.clear();
    for (auto& pair : removed_robot_hash_) {
        delete pair.second;
    }
    removed_robot_hash_.clear();
    f8::BtMgr::Instance()->BtDestory(room_agent_);
    room_agent_ = nullptr;
    team_hash_.clear();
    frame_event_data = nullptr;
    PerfMonitor::Instance()->alive_count -= alive_count_;
}

void Room::Update(int delta_time)
{
    xtimer.Update();
    if (IsGameOver() && GetFrameNo() - game_over_frameno_ > SERVER_FRAME_RATE * 20) {
        elapsed_time_ += delta_time;
        while (elapsed_time_ >= 50) {
            ++frameno_;
            elapsed_time_ -= 50;
        }
        return;
    }

    elapsed_time_ += delta_time;
    while (elapsed_time_ >= 50) {
#ifdef DEBUG1
        long long begin_tick = a8::XGetTickCount();
        long long end_tick = a8::XGetTickCount();
#endif
        room_agent_->Exec();
        if (GetFrameNo() % 2 == 0) {
            UpdateGas();
        }
#ifdef DEBUG1
        end_tick = a8::XGetTickCount();
        if (a8::XGetTickCount() - begin_tick > 1000) {
            A8_ABORT();
        }
        begin_tick = a8::XGetTickCount();
#endif
        for (auto& pair : moveable_hash_) {
            pair.second->Update(50);
        }
        for (auto& pair : task_hash_) {
            (*pair.second)(nullptr);
        }
#ifdef DEBUG1
        end_tick = a8::XGetTickCount();
        if (a8::XGetTickCount() - begin_tick > 1000) {
            A8_ABORT();
        }
        begin_tick = a8::XGetTickCount();
#endif
        if (GetFrameNo() % 2 == 0) {
            SyncFrameData();
        }
#ifdef DEBUG1
        end_tick = a8::XGetTickCount();
        if (a8::XGetTickCount() - begin_tick > 1000) {
            A8_ABORT();
        }
#endif
        ++frameno_;
        elapsed_time_ -= 50;
    }
}

bool Room::IsPveRoom()
{
    return GetMapMeta()->IsPveMap();
}

int Room::GetPlayerNum()
{
    return accountid_hash_.size();
}

int Room::GetHumanNum()
{
    return human_hash_.size();
}

Player* Room::GetPlayerByAccountId(const std::string& accountid)
{
    auto itr = accountid_hash_.find(accountid);
    return itr != accountid_hash_.end() ? itr->second : nullptr;
}

Player* Room::GetPlayerByUniId(int uniid)
{
    Entity* entity = GetEntityByUniId(uniid);
    if (!entity) {
        return nullptr;
    }
    return entity->IsEntityType(ET_Player) && entity->IsEntitySubType(EST_Player) ?
        (Player*)entity : nullptr;
}

Entity* Room::GetEntityByUniId(int uniid)
{
    auto itr = uniid_hash_.find(uniid);
    return itr != uniid_hash_.end() ? itr->second : nullptr;
}

Human* Room::GetHumanByUniId(int uniid)
{
    Entity* entity = GetEntityByUniId(uniid);
    return entity && entity->GetEntityType() == ET_Player ? (Human*)entity : nullptr;
}

int Room::AliveCount()
{
    return alive_count_;
}

Player* Room::NewPlayer()
{
    Player* hum = EntityFactory::Instance()->MakePlayer(AllocUniid());
    return hum;
}

void Room::AddPlayer(Player* hum, std::shared_ptr<BornPoint> init_born_point, bool no_matchteam)
{
    if (GetGasData().GetGasMode() != GasInactive) {
        A8_ABORT();
    }
    while (human_hash_.size() >= GetRoomMaxPlayerNum()) {
        RandRemoveAndroid();
    }
    if (init_born_point) {
        hum->born_point = init_born_point;
    } else {
        hum->born_point = AllocBornPoint(hum);
    }
    Global::Instance()->verify_set_pos = 1;
    if (!hum->born_point) {
        abort();
    } else {
        hum->SetPos(hum->born_point->RandPoint(this));
    }
    Global::Instance()->verify_set_pos = 0;
    glm::vec3 attack_dir = hum->GetPos().ToGlmVec3();
    GlmHelper::Normalize(attack_dir);
    GlmHelper::RotateY(attack_dir, a8::RandAngle());
    hum->SetAttackDir(attack_dir);
    hum->SetMoveDir(attack_dir);
    hum->room = this;
    hum->join_frameno = GetFrameNo();
    hum->Initialize();
    AddToEntityHash(hum);
    AddToMoveableHash(hum);
    AddToAccountHash(hum);
    AddToHumanHash(hum);
    AddToAliveHumanHash(hum);
    if (!no_matchteam) {
        MatchTeam(hum);
    }
#ifdef DEBUG
    hum->side = a8::RandEx(1, 2);
#endif
    hum->PushJoinRoomMsg();
    ++alive_count_;
    alive_count_chged_frameno_ = GetFrameNo();
    ++PerfMonitor::Instance()->alive_count;

    grid_service->AddCreature(hum);
    hum->FindLocation();
    hum->RefreshView();
    AddPlayerPostProc(hum);
    NotifyUiUpdate();
    OnAddHuman(hum);
    frame_event.AddEnterGame(hum->GetWeakPtrRef());
    GameLog::Instance()->GameStart(hum);
    if (custom_battle_ && custom_battle_->AllIsJoined())  {
        if (GetGasInactiveReaminTime() > 3) {
            acc_inactive_time_ += GetGasInactiveReaminTime() - 2;
        }
    }
#ifdef DEBUG
    if (IsMobaModeRoom()) {
        CreateAndroid(7);
        std::vector<Human*> humans;
        for (auto& pair : human_hash_) {
            if (pair.second->IsAndroid()) {
                humans.push_back(pair.second);
            }
        }
        if (humans.size() != 7) {
            abort();
        }
        Human* leader = hum;
        moba_team_a_ = leader->GetTeam();
        leader->InitMobaRoad();
        leader->born_point->wo_meta = std::get<0>(GetMapMeta()->moba_born_points.at(leader->side - 1));
        {
            for (size_t i = 0; i < 3; ++i) {
                Human* a = humans.at(i);
                a->side = leader->side;
                a->InitMobaRoad();
                a->born_point = leader->born_point;
                Team* t = a->GetTeam();
                leader->GetTeam()->CombineTeam(a->GetTeam());
                team_hash_.erase(t->GetTeamId());
            }
        }
        leader = humans.at(3);
        leader->side = hum->side == 1 ? 2 : 1;
        leader->InitMobaRoad();
        leader->born_point->wo_meta = std::get<0>(GetMapMeta()->moba_born_points.at(leader->side - 1));
        moba_team_b_ = leader->GetTeam();
        {
            for (size_t i = 4; i < 7; ++i) {
                Human* a = humans.at(i);
                a->side = leader->side;
                a->InitMobaRoad();
                a->born_point = leader->born_point;
                Team* t = a->GetTeam();
                leader->GetTeam()->CombineTeam(a->GetTeam());
                team_hash_.erase(t->GetTeamId());
            }
        }
    }
#endif
}

int Room::AllocUniid()
{
    while (GetEntityByUniId(++current_uniid_) ||
           later_add_hash_.find(current_uniid_) != later_add_hash_.end() ||
           current_uniid_ == 0) {}
    return current_uniid_;
}

void Room::ShuaAndroid()
{
    if (IsNewBieRoom()) {
        return;
    }
    if (IsCustomBattle()) {
        return;
    }
    if (GetGasData().GetGasMode() != GasInactive) {
        return;
    }
    int robot_num = a8::RandEx(map_meta_->refresh_robot_min_num,
                               map_meta_->refresh_robot_max_num);
    int refresh_time = a8::RandEx(map_meta_->refresh_robot_min_time,
                                  map_meta_->refresh_robot_max_time);
    if (robot_num > 0 && refresh_time > 0) {
        CreateAndroid(robot_num);
        xtimer.SetTimeoutEx
            (SERVER_FRAME_RATE * refresh_time,
             [this] (int event, const a8::Args* args)
             {
                 if (a8::TIMER_EXEC_EVENT == event) {
                     ShuaAndroid();
                 }
             },
             &xtimer_attacher_);
    }
}

void Room::ShowAndroid(Human* target, int num)
{
    if (!target->born_point) {
        return;
    }
    int i = 0;
    for (auto& pair : human_hash_) {
        Human* hum = pair.second;
        if (hum->IsAndroid() &&
            a8::HasBitFlag(hum->status, CS_Disable)) {
            if (hum->born_point) {
                DecBornPointHumanNum(hum->born_point, hum);
            }
            hum->born_point = target->born_point;
            IncBornPointHumanNum(hum->born_point, hum);
            hum->SetPos(hum->born_point->RandPoint(this));
            EnableHuman(hum);
            ++i;
            if (i >= num) {
                break;
            }
        }
    }
}

void Room::CreateAndroid(int robot_num)
{
    for (int i = 0; i < robot_num; ++i) {
        if (human_hash_.size() >= GetRoomMaxPlayerNum()) {
            return;
        }
        const mt::Robot* robot_meta = mt::Robot::RandRobot(refreshed_robot_set_);
        if (!robot_meta) {
            A8_ABORT();
        }

        Android* hum = EntityFactory::Instance()->MakeAndroid(AllocUniid());
        hum->name = robot_meta->name();
        #if 0
        hum->meta = mt::Hero::GetById(30100);
        #else
        hum->meta = mt::Hero::GetById(robot_meta->hero_id());
        #endif
        hum->robot_meta = robot_meta;
        hum->born_point = AllocBornPoint(hum);
        Global::Instance()->verify_set_pos = 1;
        if (!hum->born_point) {
            abort();
        } else {
            hum->SetPos(hum->born_point->RandPoint(this));
        }
        Global::Instance()->verify_set_pos = 0;
        glm::vec3 attack_dir = hum->GetPos().ToGlmVec3();
        GlmHelper::Normalize(attack_dir);
        GlmHelper::RotateY(attack_dir, a8::RandAngle());
        hum->SetAttackDir(attack_dir);
        hum->SetMoveDir(attack_dir);
        hum->room = this;
        hum->join_frameno = GetFrameNo();
        hum->Initialize();
        AddToEntityHash(hum);
        AddToHumanHash(hum);
        MatchTeam(hum);
        ++alive_count_;
        alive_count_chged_frameno_ = GetFrameNo();
        ++PerfMonitor::Instance()->alive_count;
        refreshed_robot_set_.insert(robot_meta->id());
        if (!CanAddToScene(hum)) {
            a8::SetBitFlag(hum->status, CS_Disable);
        } else {
            AddToAliveHumanHash(hum);
            AddToMoveableHash(hum);
            grid_service->AddCreature(hum);
            hum->FindLocation();
            hum->RefreshView();
        }
        OnAddHuman(hum);
        frame_event.AddEnterGame(hum->GetWeakPtrRef());
    }
    NotifyUiUpdate();
}

void Room::FillSMJoinedNotify(Player* self_hum, cs::SMJoinedNotify& msg)
{
    msg.set_team_mode(msg.team_mode());
    msg.set_adjust_bullet(1);
    msg.set_is_newbie_room(IsNewBieRoom() ? 1 : 0);
}

void Room::ScatterDrop(const glm::vec3& center, int drop_id, bool no_adjust)
{
    const mt::Drop* drop_meta = mt::Drop::GetById(drop_id);
    if (drop_meta) {
        std::vector<std::tuple<int, int, int>> drop_items;
        drop_meta->RandItems(drop_items);
        for (auto& item : drop_items) {
            glm::vec3 dir = GlmHelper::UP;
            GlmHelper::RotateY(dir, a8::RandAngle());
            const glm::vec3 pos = center;
            const mt::Equip* equip_meta = mt::Equip::GetById(std::get<0>(item));
            if (equip_meta) {
                #if 1
                DropItemEx(center,
                           pos,
                           std::get<0>(item),
                           std::get<1>(item),
                           std::get<2>(item),
                           true,
                           no_adjust);
                #else
                DropItemEx(center + dir * (5.0f + rand() % 50),
                           pos,
                           std::get<0>(item),
                           std::get<1>(item),
                           std::get<2>(item),
                           true,
                           no_adjust);
                #endif
            }
        }
    }
}

void Room::DropItem(const glm::vec3& pos, int item_id, int item_count, int item_lv, bool no_adjust)
{
    DropItemEx(pos, pos, item_id, item_count, item_lv, false, no_adjust);
}

void Room::DropItemEx(const glm::vec3& born_pos, const glm::vec3& pos, int item_id, int item_count, int item_lv, bool show_anim, bool no_adjust)
{
    const mt::Equip* equip_meta = mt::Equip::GetById(item_id);
    if (!equip_meta) {
        return;
    }
    int group_num = equip_meta->group_num();
    if (item_id == mt::Equip::YELLOW_STONE_ID && group_num <= 0) {
        group_num = 1;
    }
    group_num = std::max(group_num, 1);
    if (equip_meta && group_num > 0 && item_count > 0) {
#if 1
        if (equip_meta->equip_type() == EQUIP_TYPE_BULLET) {
            return;
        }
#endif
        int total_count = item_count;
        while (total_count > 0) {
            int drop_num = std::min(total_count, group_num);
            if (drop_num < 1) {
                drop_num = 1;
            }
            glm::vec3 dir = GlmHelper::UP;
            GlmHelper::RotateY(dir, a8::RandAngle());
            CreateLootEx(item_id,
                         born_pos,
                         pos + dir * (25.0f + rand() % 50),
                         drop_num,
                         item_lv,
                         show_anim,
                         no_adjust);
            total_count -= drop_num;
        }
    }
}

RoomObstacle* Room::CreateObstacle(int id, float x, float y, float z,
                                   std::shared_ptr<a8::Args> init_args)
{
    RoomObstacle* entity = InternalCreateObstacle(id, x, y, z,
                                                  [] (Obstacle*)
                                                  {
                                                  },
                                                  init_args);
    assert(entity);
    if (entity) {
        entity->BroadcastFullState(this);
    }
    return entity;
}

int Room::CreateLoot(int equip_id, const glm::vec3& born_pos, const glm::vec3& pos, int count, int equip_lv, bool no_adjust)
{
    return CreateLootEx(equip_id, born_pos, pos, count, equip_lv, false, no_adjust);
}

int Room::CreateLootEx(int equip_id, const glm::vec3& born_pos, const glm::vec3& pos, int count, int equip_lv, bool show_anim, bool no_adjust)
{
    glm::vec3 real_pos = pos;
    if (!no_adjust) {
        real_pos = born_pos;
        glm::vec3 dir = pos - born_pos;
        dir.y = 0;
        float distance = GlmHelper::Norm2D(dir);
        if (distance > 10) {
            GlmHelper::Normalize(dir);
            glm::vec3 start = born_pos;
            map_instance->Scale(start);
            for (int i = 0; i < 4; ++i) {
                glm::vec3 end = born_pos + dir * distance;
                glm::vec3 hit_point;
                bool hit_result = false;
                map_instance->Scale(end);
                bool need_adjust = true;
                if (map_instance->Raycast(start, end, hit_point, hit_result)) {
                    map_instance->UnScale(hit_point);
                    if (GlmHelper::Norm2D(hit_point - born_pos) > 5) {
                        real_pos = hit_point;
                        need_adjust = false;
                        break;
                    }
                }
                GlmHelper::RotateY(dir, glm::radians(90.0f));
            }
        }
    }
    const mt::Equip* equip_meta = mt::Equip::GetById(equip_id);
    if (equip_meta) {
        Loot* entity = EntityFactory::Instance()->MakeLoot(AllocUniid());
        entity->room = this;
        entity->meta = equip_meta;
        entity->GetMutablePos().FromGlmVec3(real_pos);
        entity->born_pos.FromGlmVec3(born_pos);
        entity->show_anim = show_anim;
        entity->create_frame_no = GetFrameNo();
        #if 1
        {
            if (entity->GetPos().GetX() >= map_meta_->map_width()) {
                entity->GetMutablePos().SetX(map_meta_->map_width() - 1);
            }
            if (entity->GetPos().GetX() < 1.0f) {
                entity->GetMutablePos().SetX(1.0f);
            }
            if (entity->GetPos().GetY() >= map_meta_->map_height()) {
                entity->GetMutablePos().SetY(map_meta_->map_height() - 1);
            }
            if (entity->GetPos().GetY() < 1.0f) {
                entity->GetMutablePos().SetY(1.0f);
            }
        }
        #endif
        entity->item_id = equip_id;
        entity->count = count;
        entity->item_level = equip_lv;
        entity->Initialize();
        AddToEntityHash(entity);
        grid_service->AddRoomEntity(this, entity);
        entity->BroadcastFullState(this);
        return entity->GetUniId();
    } else {
        return 0;
    }
}

int Room::CreateBullet(Creature* sender,
                       Creature* passenger,
                       const mt::Equip* weapon_meta,
                       const mt::Equip* bullet_meta,
                       const mt::Skill* skill_meta,
                       glm::vec3 pos,
                       glm::vec3 dir,
                       float fly_distance,
                       long long weapon_uniid,
                       int trace_target_id,
                       int hand,
                       a8::XTimerWp keep_shot_animi_timer_ptr,
                       float shot_animi_time,
                       int weapon_buff_id,
                       std::shared_ptr<std::set<int>> reporter_list)
{
    int bullet_uniid = 0;
    if (grid_service->CanAdd(pos.x, pos.z)) {
        Bullet* bullet = EntityFactory::Instance()->MakeBullet(AllocUniid());
        bullet->sender.Attach(sender);
        if (passenger) {
            bullet->passenger.Attach(passenger);
        }
        bullet->room = this;
        bullet->weapon_uniid = weapon_uniid;
        bullet->gun_meta = weapon_meta;
        bullet->meta = bullet_meta;
        bullet->skill_meta = skill_meta;
        bullet->GetMutablePos().FromGlmVec3(pos);
        bullet->dir = dir;
        bullet->born_pos.FromGlmVec3(pos);
        bullet->born_dir = dir;
        bullet->fly_distance = fly_distance;
        bullet->trace_target_id = trace_target_id;
        bullet->hand = hand;
        bullet->keep_shot_animi_timer_ptr = keep_shot_animi_timer_ptr;
        bullet->shot_animi_time = shot_animi_time;
        bullet->spec_gun_buff_id = weapon_buff_id;
        bullet->reporter_list = reporter_list;
        bullet->Initialize();
#ifdef DEBUG1
        a8::XPrintf("xxxxxxx born_pos:%f,%f,%f curr_pos:%f,%f,%f\n",
                    {
                     bullet->born_pos.x,
                     bullet->born_pos.y,
                     bullet->born_pos.z,

                     bullet->GetPos().x,
                     bullet->GetPos().y,
                     bullet->GetPos().z,

                    });
#endif
        AddObjectLater(bullet);
        bullet_uniid = bullet->GetUniId();
    }
    return bullet_uniid;
}

Car* Room::CreateCar(Human* driver,
                     int car_uniid,
                     const mt::Equip* item_meta,
                     const glm::vec3& pos,
                     int team_id,
                     std::set<int>* special_operators)
{
    Car* car = EntityFactory::Instance()->MakeCar(AllocUniid());
    car->car_uniid = car_uniid;
    car->meta = item_meta;
    car->room = this;
    car->team_id = team_id;
    car->GetMutablePos().FromGlmVec3(pos);
    car->SetAttackDir(GlmHelper::DONW);
    if (special_operators) {
        car->SetSpecialOperators(*special_operators);
    }
    car->Initialize();
    AddToEntityHash(car);
    AddToMoveableHash(car);
    grid_service->AddCreature(car);
    car->RefreshView();
    for (int buff_id : car->hero_meta_->_init_buffs) {
        car->MustBeAddBuff(car, buff_id);
    }
    return car;
}

Hero* Room::CreateHero(Creature* master,
                       const mt::Hero* meta,
                       const glm::vec3& pos,
                       const glm::vec3& dir,
                       int team_id,
                       int obj_uniid)
{
    if (obj_uniid > 0) {
        if (GetEntityByUniId(obj_uniid)) {
            abort();
        }
    } else {
        obj_uniid = AllocUniid();
    }
    Hero* hero = EntityFactory::Instance()->MakeHero(obj_uniid);
    hero->meta = meta;
    if (master) {
        hero->master.Attach(master);
    }
    hero->room = this;
    hero->GetMutablePos().FromGlmVec3(pos);
    hero->SetMoveDir(dir);
    hero->SetAttackDir(dir);
    hero->team_id = team_id;
    hero->Initialize();
    hero->hero_level = hero->GetBattleContext()->GetHeroLevel();
    AddToEntityHash(hero);
    AddToMoveableHash(hero);
    grid_service->AddCreature(hero);
    hero->RefreshView();
    for (int buff_id : meta->_init_buffs) {
        hero->MustBeAddBuff(hero, buff_id);
    }
    return hero;
}

void Room::RemoveObjectLater(RoomEntity* entity)
{
    InternalRemoveObjectLater(entity, entity->xtimer_attacher);
}

void Room::RemoveObjectLater(RoomObstacle* entity)
{
    InternalRemoveObjectLater(entity, entity->xtimer_attacher);
}

void Room::InternalRemoveObjectLater(Entity* entity, a8::Attacher& xtimer_attacher)
{
    auto remove_func =
        [this, entity] (int event, const a8::Args* args)
        {
            if (a8::TIMER_EXEC_EVENT != event) {
                return;
            }
            switch (entity->GetEntityType()) {
            case ET_Bullet:
                {
                    RemoveFromMoveableHash((Bullet*)entity);
                }
                break;
            case ET_Loot:
                {
                    entity->BroadcastDeleteState(this);
                    grid_service->DelRoomEntity(this, entity);
                }
                break;
            case ET_Player:
                {
                    RemoveFromMoveableHash((Human*)entity);
                    RemoveFromHuamnHash((Human*)entity);
                }
                break;
            case ET_Car:
                {
                    #if 0
                    RemoveFromAroundPlayers(entity->room);
                    BroadcastDeleteState(entity->room);
                    grid_service->RemoveCreature((Car*)entity);
                    #endif
                    RemoveFromMoveableHash((Car*)entity);
                }
                break;
            case ET_Hero:
                {
                    #if 0
                    BroadcastDeleteState(entity->room);
                    RemoveFromAroundPlayers(entity->room);
                    grid_service->RemoveCreature((Hero*)entity);
                    #endif
                    RemoveFromMoveableHash((Hero*)entity);
                }
                break;
            case ET_Obstacle:
                {
                    entity->BroadcastDeleteState(this);
                    grid_service->DelRoomEntity(this, entity);
                }
                break;
            default:
                {
                    A8_ABORT();
                }
                break;
            }
            RemoveFromEntityHash(entity);
#ifdef DEBUG
            #if 0
            f8::UdpLog::Instance()->Debug
                (
                 "remove object pointer:%d uniid:%d",
                 {
                  (long long)entity,
                  entity->GetUniId()
                 }
                 );
            #endif
#endif
            delete entity;
        };
    if (!entity->removing) {
        entity->removing = true;
        xtimer.SetTimeoutEx
            (NEXT_FRAME_TIMER,
             remove_func,
             &xtimer_attacher);
    }
}

void Room::OnHumanDie(Human* hum)
{
    {
        --alive_count_;
        alive_count_chged_frameno_ = GetFrameNo();
        --PerfMonitor::Instance()->alive_count;
        RemoveFromAliveHumanHash(hum);
        NotifyUiUpdate();
    }
}

void Room::OnHumanRevive(Human* hum)
{
    NotifyUiUpdate();
}

bool Room::OverBorder(const glm::vec3& pos, float radius)
{
    {
        if (pos.x - radius < 0.0001f) {
            return true;
        }
        if (pos.x + radius > map_meta_->map_width()) {
            return true;
        }
    }
    {
        if (pos.z - radius < 0.0001f) {
            return true;
        }
        if (pos.z + radius > map_meta_->map_height()) {
            return true;
        }
    }
    return false;
}

Human* Room::GetWatchWarTarget(Human* hum)
{
    if (!mt::Param::s().watchable) {
        return nullptr;
    }
    if (hum->GetTeam()) {
        Human* member = hum->GetTeam()->RandomOneAliveMember(hum);
        if (member) {
            return member;
        }
    }

    std::vector<Human*> players;
    std::vector<Human*> humans;
    for (auto& pair : human_hash_) {
        if (pair.first != hum->GetUniId() && !pair.second->dead && !a8::HasBitFlag(pair.second->status, CS_Disable)) {
            if (pair.second->IsPlayer()) {
                players.push_back(pair.second);
            } else {
                humans.push_back(pair.second);
            }
        }
    }

    Human* target = nullptr;
    if (!players.empty()) {
        target = players[rand() % players.size()];
    } else if (!humans.empty()) {
        target = humans[rand() % humans.size()];
#ifdef DEBUG1
        {
            std::sort(humans.begin(), humans.end(),
                      [hum] (Human* a, Human *b) -> bool
                      {
                          return (hum->GetPos()-a->GetPos()).Norm() < (hum->GetPos()-b->GetPos()).Norm();
                      });
            target = humans[rand() % std::min(humans.size(), (size_t)3)];
        }
#endif
    }
    if (!target) {
        target = incubator_->ActiveAndroid(hum);
    }
    return target;
}

bool Room::BattleStarted()
{
    return battle_start_frameno_ != 0 || IsPveRoom() || IsNewBieRoom();
}

int Room::GetAliveTeamNum()
{
    int num = 0;
    for (auto& pair : team_hash_) {
        if (!pair.second->IsViewTeam() && pair.second->HasAliveMember()) {
            ++num;
        }
    }
    return num;
}

void Room::TraverseTeams(std::function<bool (Team*)> cb)
{
    for (auto& pair : team_hash_) {
        if (!pair.second->IsViewTeam() && !cb(pair.second.get())) {
            break;
        }
    }
}

int Room::GetTeamNum()
{
    return team_hash_.size();
}

bool Room::CanJoin(const std::string& accountid,
                   RoomType_e self_room_type,
                   RoomMode_e self_room_mode,
                   int self_proto_version,
                   int self_channel,
                   int init_map_id,
                   const cs::CMJoin& msg)
{
    if (lock_room_) {
        return false;
    }
    if (IsNewBieRoom()) {
        return false;
    }
    if (IsCustomBattle()) {
        return false;
    }
    if (IsMobaModeRoom()) {
        return false;
    }
    if (msg.force_enter_newbie_room()) {
        return false;
    }
    if (room_mode_ != self_room_mode) {
        return false;
    }
    if (GetGasData().GetGasMode() != GasInactive) {
        return false;
    }
    if (msg.pve_instance_id() && !pve_instance){
        return false;
    }
    if (pve_instance) {
        if (pve_instance->gemini_id() != msg.pve_instance_id()) {
            return false;
        }
#if 1
        Team* team = GetAliveTeam();
        if (!team) {
            return false;
        }
        Human* hum = team->GetOneAliveMember(nullptr);
        if (!hum) {
            return false;
        }
        if (hum->team_uuid != msg.team_uuid()) {
            return false;
        }
#endif
    }
    if (init_map_id != 0) {
        auto map_instance = MapMgr::Instance()->GetMapInstance(init_map_id);
        if (map_instance && map_instance->GetMapMeta()->map_mode() == room_mode_ &&
            map_instance->map_id != map_meta_->map_id()) {
            return false;
        }
    }
    if (GetPlayerByAccountId(accountid)) {
        return false;
    }
    if (GetPlayerNum() < (int)GetRoomMaxPlayerNum() &&
        !msg.team_uuid().empty() &&
        human_hash_.size() >= GetRoomMaxPlayerNum()
        ) {
        bool have_team = false;
        for (auto& pair : human_hash_) {
            if (pair.second->team_uuid == msg.team_uuid()) {
                if (pair.second->GetTeam() && !pair.second->GetTeam()->IsFull()) {
                    have_team = true;
                    break;
                }
            }
        }//end for human_hash_
        if (have_team) {
            while (human_hash_.size() >= GetRoomMaxPlayerNum()) {
                RandRemoveAndroid();
                break;
            }
        }
    }
    return GetPlayerNum() < (int)GetRoomMaxPlayerNum();
}

bool Room::CanJoin(class MatchTeam* team)
{
    if (lock_room_) {
        return false;
    }
    if (GetGasData().GetGasMode() != GasInactive) {
        return false;
    }
    #if 0
    if (map_instance->map_id != team->GetMapId()) {
        return false;
    }
    #endif
    if (team->GetOwner()->msg->pve_instance_id() && !pve_instance){
        return false;
    }
    if (pve_instance) {
        if (pve_instance->gemini_id() != team->GetOwner()->msg->pve_instance_id()) {
            return false;
        }
#if 1
        return false;
#endif
    }
#if 0
    if (team->GetCurrMembers().size() != GetMaxTeamNum()) {
        return false;
    }
#endif
    if (GetPlayerNum() > (int)GetRoomMaxPlayerNum()) {
        return false;
    }
    for (auto& member : team->GetCurrMembers()) {
        if (GetPlayerByAccountId(member->msg->account_id())) {
            return false;
        }
    }
    return true;
}

void Room::OnPlayerOffline(Player* hum)
{
    if (GetOnlinePlayerNum() <= 0 && game_over_timer.expired()) {
        game_over_timer = xtimer.SetTimeoutWpEx
            (
             SERVER_FRAME_RATE * 15,
             [this] (int event, const a8::Args* args)
             {
                 if (a8::TIMER_EXEC_EVENT == event) {
                     if (GetOnlinePlayerNum() <= 0 && !added_to_over_room) {
                         ForceOver();
                     }
                 }
             },
             &xtimer_attacher_);
    }
}

void Room::FillSMUiUpdate(cs::SMUiUpdate& msg)
{
    for (auto& pair : car_hash_) {
        if (!pair.second.taken) {
            auto car = msg.add_car_list();
            car->set_car_id(pair.second.car_id);
            TypeConvert::ToPb(pair.second.pos, car->mutable_pos());
        }
    }
}

Team* Room::GetAliveTeam()
{
    for (auto& pair : team_hash_) {
        if (!pair.second->IsViewTeam() && pair.second->HasAliveMember()) {
            return pair.second.get();
        }
    }
    return nullptr;
}

std::shared_ptr<Team> Room::NewTeam()
{
    auto team = std::make_shared<Team>();
    team->room = this;
    team->SetTeamId(++current_teamid_);
    team_hash_[team->GetTeamId()] = team;
    return team;
}

void Room::TraversePlayerList(std::function<void (Player*)> func)
{
    for (auto& pair : accountid_hash_) {
        if (pair.second) {
            func(pair.second);
        }
    }
}

void Room::TraverseHumanList(std::function<bool (Human*)> func)
{
    for (auto& pair : human_hash_) {
        if (pair.second) {
            if (!func(pair.second)) {
                break;
            }
        }
    }
}

void Room::TraverseAliveHumanList(std::function<bool (Human*)> func)
{
    for (auto& pair : alive_human_hash_) {
        if (pair.second) {
            if (!func(pair.second)) {
                break;
            }
        }
    }
}

void Room::TraverseEntityList(std::function<bool (Entity*)> func)
{
    for (auto& pair : uniid_hash_) {
        if (pair.second) {
            if (!func(pair.second)) {
                break;
            }
        }
    }
}

void Room::TraverseAlivePlayers(std::function<bool (Human*)> func)
{
    for (auto& pair : alive_player_hash_) {
        if (!func(pair.second)) {
            break;
        }
    }
}

void Room::BroadcastDebugMsg(const std::string& debug_msg)
{
    #ifdef DEBUG
    TraverseHumanList
        (
         [debug_msg] (Human* hum) -> bool
         {
             hum->SendDebugMsg(debug_msg);
             return true;
         });
    #endif
}

void Room::UpdateGas()
{
    switch (GetGasData().GetGasMode()) {
    case GasInactive:
        UpdateGasInactive();
        break;
    case GasJump:
        UpdateGasJump();
        break;
    case GasWaiting:
        UpdateGasWaiting();
        break;
    case GasMoving:
        UpdateGasMoving();
        break;
    }
    if (GetGasData().GetGasMode() != GasInactive && GetGasData().GetGasMode() != GasJump) {
        if (GetGasData().GetGasMode() != GasInactive) {
            if (IsNewBieRoom()) {
            } else if (IsPveRoom()) {
                if (!IsGameOver() &&
                    (
                     pve_data.pve_kill_boss ||
                     IsAllRealDead() ||
                     incubator_->IsTimeOut()
                     )) {
                    game_over_ = true;
                    game_over_frameno_ = GetFrameNo();
                    OnGameOver();
                }
            } else {
                if (!IsGameOver() &&
                    alive_count_ <= GetMaxTeamNum() &&
                    GetAliveTeamNum() <= 1) {
                    game_over_ = true;
                    game_over_frameno_ = GetFrameNo();
                    OnGameOver();
                }
            }
            for (auto& pair : moveable_hash_) {
                if (!pair.second->IsCreature(this)) {
                    continue;
                }
                Creature* c = (Creature*)pair.second;
                if (c->real_dead ||
                    a8::HasBitFlag(c->status, CS_Disable)) {
                    continue;
                }
                bool b1 = a8::CircleContainCircle
                    (gas_data_.pos_old,
                     gas_data_.gas_progress,
                     glm::vec2(c->GetPos().GetX(), c->GetPos().GetZ()),
                     c->GetRadius()
                     );
                bool b2 = a8::CircleContainCircle
                    (gas_data_.pos_new,
                     gas_data_.rad_new,
                     glm::vec2(c->GetPos().GetX(), c->GetPos().GetZ()),
                     c->GetRadius()
                     );
                if (!b1 && !b2) {
                    c->poisoning = true;
                    if (!c->GetBuffByEffectId(kBET_Poisoning)) {
                        c->TryAddBuff(c, kPoisioningBuffId);
                    }
                } else {
                    c->poisoning = false;
                    c->poisoning_time = false;
                    if (c->GetBuffByEffectId(kBET_Poisoning)) {
                        c->RemoveBuffByEffectId(kBET_Poisoning);
                    }
                }
            }
        }
    }
}

void Room::UpdateGasInactive()
{
    if (IsNewBieRoom()) {
        UpdateGasInactiveNewBie();
    } else if (IsPveRoom()) {
        UpdateGasInactivePve();
    } else if (IsMobaModeRoom()) {
        UpdateGasInactiveMoba();
    } else {
        UpdateGasInactivePvp();
    }
}

void Room::UpdateGasInactivePvp()
{
    bool over = GetFrameNo() - gas_data_.gas_start_frameno >=
        GetGasInactiveTime() * SERVER_FRAME_RATE;
    if (over) {
        long long begin_tick = a8::XGetTickCount();
        gas_data_.SetGasMode(GasJump);
        gas_data_.gas_start_frameno = GetFrameNo();
        if (human_hash_.size() < GetRoomMaxPlayerNum()) {
            if (IsCustomBattle()) {
                FillCustomBattleHuman();
            } else {
                CreateAndroid(GetRoomMaxPlayerNum() - human_hash_.size());
            }
            NotifyUiUpdate();
        }
        battle_start_frameno_ = GetFrameNo();
        CombineTeam();
        FillTeam();
        ClearPostBattleAutoFreeList();
        OnBattleStart();
        InitAndroidAI();
        if (IsSandTableRoom()) {
            {
                const mt::SafeArea* first_safearea = mt::SafeArea::GetByType
                    (map_meta_->RandSafeArea());
                gas_data_.SetGasMode(GasWaiting);
                gas_data_.old_area_meta = first_safearea;
                gas_data_.new_area_meta = mt::SafeArea::GetById(first_safearea->id() + 1);
                gas_data_.pos_list = mt::SafeAreaPos::RandSafeAreaPos(first_safearea->id());
                gas_data_.gas_progress = gas_data_.old_area_meta->rad();
                gas_data_.gas_start_frameno = GetFrameNo();
#if 1
                gas_data_.pos_old = glm::vec2(first_safearea->x1(),
                                              first_safearea->y1());
#else
                gas_data_.pos_old = map_meta_->first_safearea_center;
#endif
                gas_data_.pos_old_bk = gas_data_.pos_old;
                {
                    if (gas_data_.pos_list && !gas_data_.pos_list->_poses.empty()) {
                        gas_data_.pos_old = glm::vec2(
                                                      gas_data_.pos_list->_poses[0].x,
                                                      gas_data_.pos_list->_poses[0].z
                                                      );
                        gas_data_.pos_old_bk = gas_data_.pos_old;
                    }
                    bool gen_ok = GenSmallCircle();
                    if (!gen_ok) {
                        A8_ABORT();
                    }
                }
                gas_data_.rad_old = gas_data_.old_area_meta->rad();
                gas_data_.rad_new = gas_data_.new_area_meta->rad();
                if (map_meta_->init_gas_ring() > 0) {
                    ForwardGasRing(map_meta_->init_gas_ring());
                }
            }
            TraverseHumanList
                (
                 [] (Human* hum) -> bool
                 {
                     hum->RemoveBuffByEffectId(kBET_ThroughWall);
                     hum->OnLand();
                     if (hum->GetEntitySubType() == EST_Player) {
                         GameLog::Instance()->GameStart((Player*)hum);
                     }
                     return true;
                 });
            NotifyUiUpdate();
            NotifyGameStart();
            sand_table_->OnGameStart();
            if (custom_battle_) {

            }
        } else {
            ShuaPlane();
            RoomMgr::Instance()->ActiveRoom(GetRoomUuid());
#if 1
            int auto_jump_interval = mt::Param::GetIntParam("auto_jump_interval");
            auto_jump_timer_ = xtimer.SetIntervalWpEx
                (SERVER_FRAME_RATE * auto_jump_interval + SERVER_FRAME_RATE * 3,
                 [this] (int event, const a8::Args* args)
                 {
                     if (a8::TIMER_EXEC_EVENT == event) {
                         AutoJump();
                     }
                 },
                 &xtimer_attacher_);
#else
            int auto_jump_interval_delay = mt::Param::GetIntParam("auto_jump_interval_delay");
            xtimer.SetTimeoutEx
                (
                 SERVER_FRAME_RATE * auto_jump_interval_delay,
                 [this] (int event, const a8::Args* args)
                 {
                     int auto_jump_interval = mt::Param::GetIntParam("auto_jump_interval");
                     auto_jump_timer_ = xtimer.SetIntervalWpEx
                         (SERVER_FRAME_RATE * auto_jump_interval,
                          [this] (int event, const a8::Args* args)
                          {
                              if (a8::TIMER_EXEC_EVENT == event) {
                                  AutoJump();
                              }
                          },
                          &xtimer_attacher_);
                 },
                 &xtimer_attacher_);
#endif
        }
#ifdef DEBUG
            a8::XPrintf("max_run_delay:%d %d\n", {PerfMonitor::Instance()->max_run_delay_time,
                        a8::XGetTickCount() - begin_tick});
#endif
    }
}

void Room::UpdateGasInactivePve()
{
    if (GetFrameNo() - gas_data_.gas_start_frameno >=
        GetGasInactiveTime() * SERVER_FRAME_RATE) {
        {
            auto first_safearea = pve_mode_meta->_area[0];
            gas_data_.SetGasMode(GasWaiting);
            gas_data_.gas_start_frameno = GetFrameNo();
            battle_start_frameno_ = GetFrameNo();
            gas_data_.old_area_meta = first_safearea;
            gas_data_.new_area_meta = pve_mode_meta->_area[0];
            gas_data_.gas_start_frameno = GetFrameNo();
            gas_data_.pos_new = glm::vec2(gas_data_.new_area_meta->x1(),
                                          gas_data_.new_area_meta->y1());
            gas_data_.pos_old = gas_data_.pos_new;
            gas_data_.pos_old_bk = gas_data_.pos_old;
            gas_data_.rad_old = gas_data_.old_area_meta->rad();
            gas_data_.rad_new = gas_data_.new_area_meta->rad();
            gas_data_.gas_progress = 0;
        }
        NotifyUiUpdate();
        ClearPostBattleAutoFreeList();
        OnBattleStart();
        NotifyGameStart();
        InitAndroidAI();
        incubator_->InitPve();
    }
}

void Room::UpdateGasInactiveNewBie()
{
    {
        auto first_safearea = mt::Param::s().newbie_room_area.at(0);
        gas_data_.SetGasMode(GasWaiting);
        gas_data_.gas_start_frameno = GetFrameNo();
        battle_start_frameno_ = GetFrameNo();
        gas_data_.old_area_meta = first_safearea;
        gas_data_.new_area_meta = mt::Param::s().newbie_room_area.at(0);
        gas_data_.gas_start_frameno = GetFrameNo();
        gas_data_.pos_new = glm::vec2(gas_data_.new_area_meta->x1(),
                                      gas_data_.new_area_meta->y1());
        gas_data_.pos_old = gas_data_.pos_new;
        gas_data_.pos_old_bk = gas_data_.pos_old;
        gas_data_.rad_old = gas_data_.old_area_meta->rad();
        gas_data_.rad_new = gas_data_.new_area_meta->rad();
        gas_data_.gas_progress = 0;
    }
    NotifyUiUpdate();
    ClearPostBattleAutoFreeList();
    OnBattleStart();
    NotifyGameStart();
    InitAndroidAI();
}

void Room::UpdateGasInactiveMoba()
{
    if (GetFrameNo() - gas_data_.gas_start_frameno >=
        GetGasInactiveTime() * SERVER_FRAME_RATE) {
        {
            const mt::SafeArea* first_safearea = mt::SafeArea::GetByType
                (map_meta_->RandSafeArea());
            gas_data_.SetGasMode(GasWaiting);
            gas_data_.gas_start_frameno = GetFrameNo();
            battle_start_frameno_ = GetFrameNo();
            gas_data_.old_area_meta = first_safearea;
            gas_data_.new_area_meta = first_safearea;
            gas_data_.gas_start_frameno = GetFrameNo();
            gas_data_.pos_new = glm::vec2(gas_data_.new_area_meta->x1(),
                                          gas_data_.new_area_meta->y1());
            gas_data_.pos_old = gas_data_.pos_new;
            gas_data_.pos_old_bk = gas_data_.pos_old;
            gas_data_.rad_old = gas_data_.old_area_meta->rad();
            gas_data_.rad_new = gas_data_.new_area_meta->rad();
            gas_data_.gas_progress = 0;
        }
        TraverseHumanList
            (
             [] (Human* hum) -> bool
             {
                 hum->RemoveBuffByEffectId(kBET_ThroughWall);
                 hum->OnLand();
                 if (hum->GetEntitySubType() == EST_Player) {
                     GameLog::Instance()->GameStart((Player*)hum);
                 }
                 return true;
             });
        NotifyUiUpdate();
        battle_start_frameno_ = GetFrameNo();
#if 1
        Global::Instance()->verify_set_pos = 1;
        for (auto& pair : human_hash_) {
            Human* hum = pair.second;
            if (!hum->born_point) {
                abort();
            } else {
                hum->SetPos(hum->born_point->RandPoint(this));
            }
            a8::XPrintf("set_pos side:%d %d pos:%f %f %f\n", {
                    hum->side,
                    hum->GetUniId(),
                        hum->GetPos().GetX(),
                        hum->GetPos().GetY(),
                        hum->GetPos().GetZ(),
                });
            hum->GetMovement()->ClearPath();
            grid_service->MoveCreature(hum);
        }
        Global::Instance()->verify_set_pos = 0;
#else
        CombineTeam();
        FillTeam();
#endif
        ClearPostBattleAutoFreeList();
        OnBattleStart();
        NotifyGameStart();
        InitAndroidAI();
        sand_table_->OnGameStart();
    }
}

void Room::UpdateGasWaiting()
{
    if (IsPveRoom() || IsNewBieRoom() || IsMobaModeRoom()) {

    } else {
        if (GetFrameNo() - gas_data_.gas_start_frameno >=
            gas_data_.old_area_meta->wait_time() * SERVER_FRAME_RATE) {
            gas_data_.SetGasMode(GasMoving);
            gas_data_.gas_start_frameno = GetFrameNo();;
        }
    }
}

void Room::UpdateGasMoving()
{
    if (GetFrameNo() - gas_data_.gas_start_frameno > 0 && !gas_data_.is_last_gas) {
        float distance = gas_data_.old_area_meta->shrink_speed() *
            ((GetFrameNo() - gas_data_.gas_start_frameno) * (1.0 / SERVER_FRAME_RATE));
        gas_data_.gas_progress = std::max(gas_data_.rad_old - distance, gas_data_.rad_new);
        if (!(gas_data_.pos_new == gas_data_.pos_old)) {
            glm::vec2 p1 = gas_data_.pos_new - gas_data_.pos_old_bk;
            gas_data_.pre_pos_old = gas_data_.pos_old;
            if (glm::length(p1) - distance <= 0.01f) {
                gas_data_.pos_old = gas_data_.pos_new;
            } else {
                glm::vec2 dir = p1;
                dir = glm::normalize(dir);
                gas_data_.pos_old = gas_data_.pos_old_bk + dir * distance;
            }
            if (gas_data_.rad_old - distance <= gas_data_.rad_new) {
                #if 1
                assert(gas_data_.rad_old - distance <= gas_data_.rad_new + 5);
                #else
                assert(gas_data_.pos_new == gas_data_.pos_old);
                #endif
            }
        }
        if (std::abs(gas_data_.gas_progress - gas_data_.rad_new) <= 0.001f) {
            if (gas_data_.new_area_meta->IsLastGas()) {
                gas_data_.is_last_gas = true;
                return;
            }
            ForwardGasRing(1);
        }
    }
}

void Room::UpdateGasJump()
{

    glm::vec3 len_vec = plane.dir *
        ((GetFrameNo() - GetGasData().gas_start_frameno)*airline_->plane_speed() / SERVER_FRAME_RATE);
    plane.curr_pos = plane.start_point + len_vec;
    if (GlmHelper::Norm(plane.end_point - plane.start_point) <=
        GlmHelper::Norm(len_vec) + mt::Param::s().early_parachute_jump) {
        TraverseHumanList
            (
             [] (Human* hum) -> bool
             {
                 if (hum->HasBuffEffect(kBET_Fly)) {
                     hum->DoJump();
                 }
                 if (hum->GetEntitySubType() == EST_Player) {
                     GameLog::Instance()->GameStart((Player*)hum);
                 }
                 return true;
             });
        const mt::SafeArea* first_safearea = mt::SafeArea::GetByType
            (map_meta_->RandSafeArea());
        gas_data_.SetGasMode(GasWaiting);
        gas_data_.old_area_meta = first_safearea;
        gas_data_.new_area_meta = mt::SafeArea::GetById(first_safearea->id() + 1);
        gas_data_.pos_list = mt::SafeAreaPos::RandSafeAreaPos(first_safearea->id());
        gas_data_.gas_progress = gas_data_.old_area_meta->rad();
        gas_data_.gas_start_frameno = GetFrameNo();
        #if 1
        gas_data_.pos_old = glm::vec2(first_safearea->x1(),
                                      first_safearea->y1());
        #else
        gas_data_.pos_old = map_meta_->first_safearea_center;
        #endif
        gas_data_.pos_old_bk = gas_data_.pos_old;
        {
            if (gas_data_.pos_list && !gas_data_.pos_list->_poses.empty()) {
                gas_data_.pos_old = glm::vec2(
                                              gas_data_.pos_list->_poses[0].x,
                                              gas_data_.pos_list->_poses[0].z
                                              );
                gas_data_.pos_old_bk = gas_data_.pos_old;
            }
            bool gen_ok = GenSmallCircle();
            if (!gen_ok) {
                A8_ABORT();
            }
        }
        gas_data_.rad_old = gas_data_.old_area_meta->rad();
        gas_data_.rad_new = gas_data_.new_area_meta->rad();
        if (map_meta_->init_gas_ring() > 0) {
            ForwardGasRing(map_meta_->init_gas_ring());
        }
        if (!auto_jump_timer_.expired()) {
            xtimer.Delete(auto_jump_timer_);
        }
        air_drop_->Init();
        air_raid_->Init();
        NotifyGameStart();
        NotifyUiUpdate();
    }
}

bool Room::GenSmallCircle()
{
    if (gas_data_.new_area_meta->rad() >= gas_data_.old_area_meta->rad()) {
        A8_ABORT();
    }
    if (gas_data_.pos_list && gas_data_.gas_count + 1 < gas_data_.pos_list->_poses.size()) {
        gas_data_.pos_new =
            glm::vec2(
                      gas_data_.pos_list->_poses[gas_data_.gas_count + 1].x,
                      gas_data_.pos_list->_poses[gas_data_.gas_count + 1].z
                      );
    } else {
        gas_data_.pos_new = glm::vec2(gas_data_.new_area_meta->x1(),
                                      gas_data_.new_area_meta->y1());
    }
    return true;
}

void Room::MatchTeam(Human* hum)
{
    if (!hum->team_uuid.empty()) {
        for (auto& pair : human_hash_) {
            if (pair.second != hum) {
                if (pair.second->team_uuid == hum->team_uuid) {
                    if (pair.second->GetTeam() && !pair.second->GetTeam()->IsFull()) {
                        pair.second->GetTeam()->AddMember(hum);
                        return;
                    }
                }
            }
        }//end for human_hash_
        if (mt::Param::s().prebattle_combine_team) {
            if (!hum->GetTeam() && hum->auto_fill && combineable_team_hash_.size() > 1) {
                for (auto& pair : combineable_team_hash_) {
                    auto team = pair.second;
                    if (team->CanCombine(hum)) {
                        team->AddMember(hum);
                        team->AddCombineMemberNum(hum->init_team_member_num);
                        return;
                    }
                }
            }
        }//end if
    }

    if (!hum->GetTeam()) {
        auto new_team = NewTeam();
        new_team->SetInitTeamMemberNum(hum->init_team_member_num);
        new_team->SetAutoFill(hum->auto_fill);
        new_team->AddMember(hum);
        if (hum->IsPlayer() && hum->auto_fill) {
            combineable_team_hash_[new_team->GetTeamId()] = new_team;
        }
    }
}

void Room::CombineTeam()
{
    if (!mt::Param::s().prebattle_combine_team) {
        return;
    }
    std::map<int, size_t> need_combine_teams;
    std::map<int, size_t> need_combine_teams_copy;

    int first_team_id = 0;
    int total_count = 0;
    for (auto& pair : team_hash_) {
        auto team = pair.second;
        if (!team->AllIsRunAway()) {
            team->TraverseMembers
                (
                 [this, team, &first_team_id, &total_count, &need_combine_teams] (Human* member) -> bool
                 {
                     if (member->auto_fill && !member->team_uuid.empty()) {
                         if (first_team_id == 0) {
                             first_team_id = team->GetTeamId();
                         }
                         need_combine_teams[team->GetTeamId()] = team->GetMemberNum();
                         total_count += team->GetMemberNum();
                         return false;
                     }
                     return true;
                 });
        }
    }

    if (total_count <= 1) {
        return;
    }

    if (need_combine_teams.size() <= 2) {
        return;
    }
    if (total_count == 3) {
        int del_team_id = 0;
        for (auto& pair : need_combine_teams) {
            if (pair.first != first_team_id) {
                del_team_id = pair.first;
                break;
            }
        }
        if (del_team_id != 0) {
            auto itr = need_combine_teams.find(del_team_id);
            if (itr != need_combine_teams.end()) {
                total_count -= itr->second;
                need_combine_teams.erase(itr);
            }
        }
        if (total_count <= 1) {
            abort();
        }
    }

    size_t first_team_num = 4;
    switch (total_count) {
    case 3:
        {
            first_team_num = 2;
            break;
        }
    case 4:
        {
            first_team_num = 2;
            break;
        }
    case 5:
        {
            first_team_num = 3;
            break;
        }
    default:
        {
            first_team_num = 4;
            break;
        }
    }

    need_combine_teams_copy = need_combine_teams;
    int try_count = 0;
    do {
        for (auto& pair1 : need_combine_teams) {
            bool combine_ok = false;
            for (auto& pair2 : need_combine_teams_copy) {
                int team_id1 = pair1.first;
                int team_id2 = pair2.first;
                if (team_id2 == first_team_id) {
                    continue;
                }
                if (team_id1 == first_team_id) {
                    if (pair1.second + pair2.second > first_team_num) {
                        continue;
                    }
                }
                if (team_id1 != team_id2) {
                    if (pair1.second + pair2.second <= GetMaxTeamNum()) {
                        int new_team_num = pair1.second + pair2.second;
                        {
                            std::shared_ptr<Team> team1 = team_hash_[team_id1];
                            std::shared_ptr<Team> team2 = team_hash_[team_id2];
                            if (team1->GetMemberNum() + team2->GetMemberNum() > GetMaxTeamNum()) {
                                f8::UdpLog::Instance()->Warning("team_member > 4 :%d",
                                                                {
                                                                 team1->GetMemberNum() + team2->GetMemberNum()
                                                                });
                            }
                            if (pair1.first == first_team_id || pair1.second >= pair2.second) {
                                team1->CombineTeam(team2.get());
                                team_hash_.erase(pair2.first);
                            } else {
                                team2->CombineTeam(team1.get());
                                team_hash_.erase(pair1.first);
                            }
                        }

                        if (first_team_id == team_id1) {
                            if (pair1.second + pair2.second >= first_team_num) {
                                need_combine_teams.erase(team_id1);
                                need_combine_teams_copy.erase(team_id1);
                            } else {
                                pair1.second = new_team_num;
                                if (need_combine_teams_copy.find(pair1.first) !=
                                    need_combine_teams_copy.end()) {
                                    need_combine_teams_copy[pair1.first] = new_team_num;
                                }
                            }
                            need_combine_teams.erase(team_id2);
                            need_combine_teams_copy.erase(team_id2);
                        } else {
                            if (pair1.second >= pair2.second) {
                                if (pair1.second + pair2.second >= GetMaxTeamNum()) {
                                    need_combine_teams.erase(team_id1);
                                    need_combine_teams_copy.erase(team_id1);
                                } else {
                                    pair1.second = new_team_num;
                                    if (need_combine_teams_copy.find(pair1.first) !=
                                        need_combine_teams_copy.end()) {
                                        need_combine_teams_copy[pair1.first] = new_team_num;
                                    }
                                }
                                need_combine_teams.erase(team_id2);
                                need_combine_teams_copy.erase(team_id2);
                            } else {
                                if (pair1.second + pair2.second >= GetMaxTeamNum()) {
                                    need_combine_teams.erase(team_id2);
                                    need_combine_teams_copy.erase(team_id2);
                                } else {
                                    pair2.second = new_team_num;
                                    if (need_combine_teams.find(pair2.first) !=
                                        need_combine_teams.end()) {
                                        need_combine_teams[pair2.first] = new_team_num;
                                    }
                                }
                                need_combine_teams.erase(team_id1);
                                need_combine_teams_copy.erase(team_id1);
                            }
                        }
                        combine_ok = true;
                        break;
                    }
                }
            }//end for
            if (combine_ok) {
                break;
            }
        }
    } while (++try_count < 20);
}

void Room::FillTeam()
{
    std::vector<std::shared_ptr<Team>> free_team_list;
    for (auto& pair : team_hash_) {
        if (pair.second->IsFreeTeam()) {
            free_team_list.push_back(pair.second);
        }
    }
    std::random_shuffle(free_team_list.begin(), free_team_list.end());
    for (auto& pair : accountid_hash_) {
        if (free_team_list.empty()) {
            break;
        }
        if (!pair.second->GetTeam()->IsFull()) {
            for (int i = pair.second->GetTeam()->GetMemberNum(); i < GetMaxTeamNum(); ++i) {
                if (!free_team_list.empty()) {
                    std::shared_ptr<Team> b_team = free_team_list.at(free_team_list.size() - 1);
                    pair.second->GetTeam()->CombineTeam(b_team.get());
                    free_team_list.erase(free_team_list.begin() + free_team_list.size() - 1);
                    team_hash_.erase(b_team->GetTeamId());
                }
            }
        }
    }
    while (free_team_list.size() > 1) {
        std::shared_ptr<Team> a_team = free_team_list[0];
        free_team_list.erase(free_team_list.begin());
        for (int i = 1; i < 4; ++i) {
            if (!free_team_list.empty()) {
                std::shared_ptr<Team> b_team = free_team_list.at(0);
                a_team->CombineTeam(b_team.get());
                free_team_list.erase(free_team_list.begin());
                team_hash_.erase(b_team->GetTeamId());
            }
        }
    }
    for (auto& pair : team_hash_) {
        pair.second->GenBattleUuid();
        if (pair.second->HasPlayer()) {
            batch_sync_->AddTeam(pair.second.get());
        }
    }
}

void Room::ShuaPlane()
{
    airline_ = mt::AirLine::RandAirLine(map_meta_->map_id());
    if (!airline_) {
        A8_ABORT();
    }
    plane.start_point = airline_->GetStartPoint();
    plane.end_point = airline_->GetEndPoint();
    plane.dir = plane.end_point - plane.start_point;
    GlmHelper::Normalize(plane.dir);
    plane.curr_pos = plane.start_point;
    last_player_jump_pos = plane.curr_pos;

    for (auto& pair : human_hash_) {
        pair.second->MustBeAddBuff(pair.second, FLY_BUFFID);
        pair.second->ResetAllSkillCd();
        Global::Instance()->verify_set_pos = 1;
        pair.second->GetMutablePos().FromGlmVec3(plane.curr_pos);
        Global::Instance()->verify_set_pos = 0;
        pair.second->GetMovement()->ClearPath();
        pair.second->SetAttackDir(plane.dir);
        pair.second->SetMoveDir(plane.dir);
        grid_service->MoveCreature(pair.second);
        pair.second->AddToNewObjects(pair.second);
        pair.second->ActiveAllSkill();
    }
}

RoomObstacle* Room::InternalCreateObstacle(int id, float x, float y, float z,
                                           std::function<void (Obstacle*)> on_precreate,
                                           std::shared_ptr<a8::Args> init_args)
{
    const mt::MapThing* thing = mt::MapThing::GetById(id);
    if (thing) {
        RoomObstacle* entity = EntityFactory::Instance()->MakeRoomObstacle(AllocUniid());
        entity->meta = thing;
        entity->room = this;
        entity->init_args = init_args;
        {
            Position pos;
            pos.SetX(x);
            pos.SetY(y);
            pos.SetZ(z);
            entity->SetPos(pos);
        }
        entity->Initialize();
        if (on_precreate) {
            on_precreate(entity);
        }
        AddToEntityHash(entity);
        grid_service->AddRoomEntity(this, entity);
        return entity;
    }
    return nullptr;
}

void Room::AddObjectLater(RoomEntity* entity)
{
    auto add_func =
        [entity] (int event, const a8::Args* args)
        {
            if (a8::TIMER_EXEC_EVENT == event) {
                if (entity->IsEntityType(ET_Bullet)) {
                    MoveableEntity* moveableentity = (MoveableEntity*)entity;
                    entity->room->AddToMoveableHash(moveableentity);

                }
                entity->room->AddToEntityHash(entity);
                entity->room->RemoveFromLaterAddHash(entity);
            }
        };
    AddToLaterAddHash(entity);
    xtimer.SetTimeoutEx
        (0,
         add_func,
         &entity->xtimer_attacher);
}

void Room::OnGameOver()
{
    for (auto& pair : human_hash_) {
        pair.second->SendGameOver();
    }
    int param1 = 1;
    xtimer.SetIntervalEx
        (
         SERVER_FRAME_RATE * 6,
         [this, param1] (int event, const a8::Args* args) mutable
         {
             if (a8::TIMER_EXEC_EVENT == event) {
                 if (GetOnlinePlayerNum() <= 0 && !added_to_over_room) {
                     RoomMgr::Instance()->AddOverRoom(room_uuid_);
                     xtimer.DeleteCurrentTimer();
                     return;
                 }
#ifdef DEBUG
                 if (param1 > 10) {
#else
                 if (param1 > 5) {
#endif
                     RoomMgr::Instance()->AddOverRoom(room_uuid_);
                 }
                 for (auto& pair : human_hash_) {
                     pair.second->SendGameOver();
                 }
                 ++param1;
             }
         },
         &xtimer_attacher_);
}

void Room::RandRemoveAndroid()
{
    Human* hum = nullptr;
    if (!hum) {
        for (auto& pair : human_hash_) {
            if (pair.second->IsAndroid() && pair.second->team_uuid.empty()) {
                hum = pair.second;
                break;
            }
        }
    }
    if (hum) {
        if (hum->team_id != 0) {
            team_hash_.erase(hum->team_id);
        }
        if (hum->born_point) {
            DecBornPointHumanNum(hum->born_point, hum);
        }
        grid_service->DeatchHuman(hum);
        RemoveFromMoveableHash(hum);
        RemoveFromEntityHash(hum);
        RemoveFromHuamnHash(hum);
        RemoveFromAliveHumanHash(hum);
        AddToRemovedRobotHash(hum);
        --alive_count_;
        alive_count_chged_frameno_ = GetFrameNo();
        --PerfMonitor::Instance()->alive_count;
        for (auto& pair : human_hash_) {
            pair.second->RemovePartObjects(hum);
        }
        frame_event.AddExitGame(hum->GetWeakPtrRef());
    }
}

void Room::NotifyUiUpdate()
{
    xtimer.SetTimeoutEx
        (0,
         [this] (int event, const a8::Args* args)
         {
             if (a8::TIMER_EXEC_EVENT == event) {
                 TraversePlayerList(
                                    [] (Player * hum)
                                    {
                                        hum->SendUIUpdate();
                                    });
             }
         },
         &xtimer_attacher_);
}

void Room::UpdateCarObject(int old_uniid, int new_uniid, const glm::vec3 pos)
{
    auto itr = car_hash_.find(old_uniid);
    if (itr != car_hash_.end()) {
        CarObject new_obj = itr->second;
        new_obj.pos = pos;
        car_hash_[new_uniid] = new_obj;
        car_hash_.erase(old_uniid);
    } else {
        assert(false);
    }
}

void Room::TakeOnCarObject(int car_uniid)
{
    auto itr = car_hash_.find(car_uniid);
    if (itr != car_hash_.end()) {
        itr->second.taken = true;
    }
}

void Room::TakeOffCarObject(int car_uniid, glm::vec3 pos)
{
    auto itr = car_hash_.find(car_uniid);
    if (itr != car_hash_.end()) {
        itr->second.pos = pos;
        itr->second.taken = false;
    }
}

int Room::CreateAndTakeonCar(int car_id, glm::vec3 pos)
{
    const mt::Equip* item_meta = mt::Equip::GetById(car_id);
    if (!item_meta) {
        return -1;
    }
    int car_uniid = AllocUniid();
    CarObject car;
    car.car_id = car_id;
    car.pos = pos;
    car.taken = true;
    car_hash_[car_uniid] = car;
    return car_uniid;
}

bool Room::HaveMyTeam(const std::string& team_uuid)
{
    if (team_uuid.empty()) {
        return false;
    }
    for (auto& pair : accountid_hash_) {
        if (pair.second->team_uuid == team_uuid) {
            return true;
        }
    }
    return false;
}

std::shared_ptr<BornPoint> Room::AllocBornPoint(Human* hum)
{
    std::shared_ptr<BornPoint> born_point = nullptr;
    if (hum->born_point) {
        std::vector<std::shared_ptr<BornPoint>> point_list;
        std::vector<std::shared_ptr<BornPoint>> free_point_list;
        std::shared_ptr<BornPoint> pre_point = nullptr;
        std::shared_ptr<BornPoint> reserve_point = nullptr;
        for (auto& pair : born_point_hash_) {
            if (pair.second != hum->born_point) {
                if (pair.second->player_num + pair.second->android_num <
                    pair.second->GetNum()) {
                    point_list.push_back(pair.second);
                    free_point_list.push_back(pair.second);;
                }
                if (!reserve_point || rand() % 100 < 10) {
                    reserve_point = pair.second;
                }
            } else {
                pre_point = pair.second;
            }
        }
        if (!free_point_list.empty()) {
            born_point = free_point_list[rand() % free_point_list.size()];
            if (pre_point) {
                DecBornPointHumanNum(pre_point, hum);
            }
        } else {
            if (point_list.empty() && reserve_point) {
                born_point = ForceTakeBornPoint(hum, reserve_point);
            } else {
                born_point = !point_list.empty() ? point_list[rand() % point_list.size()] : nullptr;
                if (pre_point) {
                    DecBornPointHumanNum(pre_point, hum);
                }
            }
        }
    } else {
        std::vector<std::shared_ptr<BornPoint>> free_point_list;
        for (auto& pair : born_point_hash_) {
            if (pair.second->player_num + pair.second->android_num <
                pair.second->GetNum()) {
                free_point_list.push_back(pair.second);
            }
        }
        if (!free_point_list.empty()) {
            born_point = free_point_list[rand() % free_point_list.size()];
        }
    }
    if (born_point) {
        IncBornPointHumanNum(born_point, hum);
    }
#ifdef DEBUG
    if (!born_point) {
        A8_ABORT();
    }
#endif
    return born_point;
}

std::shared_ptr<BornPoint> Room::GetBornPoint(int point_uniid)
{
    auto itr = born_point_hash_.find(point_uniid);
    return itr != born_point_hash_.end() ? itr->second : nullptr;
}

void Room::CreateSpawnPoints()
{
    #if 0
    if (born_point_hash_.empty()) {
        for (size_t i = 0; i < 10; ++i) {
            int uniid = AllocUniid();
            std::shared_ptr<BornPoint> born_point = std::make_shared<BornPoint>();
            born_point_hash_[uniid] = born_point;
            if (IsPveRoom()) {
                std::shared_ptr<mt::WorldObject> wo_meta = std::make_shared<mt::WorldObject>();
                wo_meta->pos = pve_instance->_spawn_point;
                born_point->wo_meta = wo_meta;
            }
        }
    }
    #endif
}

void Room::IncBornPointHumanNum(std::shared_ptr<BornPoint> point, Human* hum)
{
    switch (hum->GetEntitySubType()) {
    case EST_Player:
        {
            ++point->player_num;
        }
        break;
    case EST_Android:
        {
            ++point->android_num;
        }
        break;
    default:
        {
        }
        break;
    }
}

void Room::DecBornPointHumanNum(std::shared_ptr<BornPoint> point, Human* hum)
{
    switch (hum->GetEntitySubType()) {
    case EST_Player:
        {
            --point->player_num;
        }
        break;
    case EST_Android:
        {
            --point->android_num;
        }
        break;
    default:
        {

        }
        break;
    }
}

void Room::SecondRandPoint()
{
    #ifdef DEBUG
    CheckPartObjects();
    #endif
    for (auto& pair : accountid_hash_) {
        Human* hum = pair.second;
        hum->born_point = AllocBornPoint(hum);
    }
    CombineTeamBornPoint();
    #ifdef DEBUG
    CheckPartObjects();
    #endif
}

void Room::NotifyGameStart()
{
    cs::SMGameStart msg;
    for (auto& pair : accountid_hash_) {
        pair.second->SendNotifyMsg(msg);
    }
    for (auto& pair : human_hash_) {
        a8::SetBitFlag(pair.second->status, CS_DisableAttack);
    }

    xtimer.SetTimeoutEx
        (mt::Param::GetIntParam("prepare_time", 3000) / FRAME_RATE_MS,
         [this] (int event, const a8::Args* args)
         {
             if (a8::TIMER_EXEC_EVENT == event) {
                 for (auto& pair : human_hash_) {
                     a8::UnSetBitFlag(pair.second->status, CS_DisableAttack);
                     for (int buff_id : map_meta_->buff_list) {
                         const mt::Buff* buff_meta = mt::Buff::GetById(buff_id);
                         if (buff_meta) {
#if 0
                             pair.second->AddBuff(pair.second,
                                                  buff_meta,
                                                  nullptr);
#endif
                         }
                     }
                 }
             }
         },
         &xtimer_attacher_);
}

long long Room::GetGasInactiveTime()
{
    long long inactive_time = 0;
    if (IsPveRoom()) {
        inactive_time = 10;
    } else {
        if (IsPvpRankModeRoom()) {
            if (IsPvpMasterRankModeRoom()) {
                inactive_time = mt::Param::s().master_rank_gas_inactive_time;
            } else {
                inactive_time = mt::Param::s().rank_gas_inactive_time;
            }
        } else {
            inactive_time = mt::Param::s().gas_inactive_time;
        }
    }
#ifdef DEBUG
    if (!f8::IsTestEnv()) {
        inactive_time = 12;
    }
#endif
    inactive_time = std::max((long long)0, inactive_time - acc_inactive_time_);
    if (IsNewerMap()) {
        inactive_time = 10;
    }
    return inactive_time;
}

long long Room::GetGasInactiveReaminTime()
{
    if (GetGasData().GetGasMode() != GasInactive) {
        return -1;
    }
    long long remain_time = GetGasInactiveTime() -
        (GetFrameNo() - GetGasData().gas_start_frameno) * FRAME_RATE_MS;
    return std::max((int)remain_time, 0);
}

void Room::EnableHuman(Human* target)
{
    if (!RuningInTimer()) {
        A8_ABORT();
    }
    #ifdef DEBUG
    CheckPartObjects();
    #if 0
    f8::UdpLog::Instance()->Debug("enablehuman %d %d",
                                  {
                                   (long long)target,
                                   target->GetUniId()
                                  });
    #endif
    #endif
    #if 0
    target->OnEnable();
    AddToMoveableHash(target);
    if (!target->real_dead) {
        AddToAliveHumanHash(target);
    }
    #else
    if (a8::HasBitFlag(target->status, CS_Disable)) {
        target->OnEnable();
        AddToMoveableHash(target);
        if (!target->real_dead) {
            AddToAliveHumanHash(target);
        }
    } else {
        A8_ABORT();
    }
    #endif
    #ifdef DEBUG
    CheckPartObjects();
    #endif
}

void Room::DisableHuman(Human* target)
{
    #if 0
    if (!RuningInTimer()) {
        A8_ABORT();
    }
    #endif
    #ifdef DEBUG
    CheckPartObjects();
    #if 0
    f8::UdpLog::Instance()->Debug("disablehuman %d %d",
                                  {
                                   (long long)target,
                                   target->GetUniId()
                                  });
    #endif
    #endif
    if (!a8::HasBitFlag(target->status, CS_Disable)) {
        target->OnDisable();
        RemoveFromMoveableHash(target);
        RemoveFromAliveHumanHash(target);
    }
    #ifdef DEBUG
    CheckPartObjects();
    #endif
}

bool Room::HasPlayerInRound(const glm::vec3& pos, float rad)
{
    for (auto& pair : accountid_hash_) {
        if (std::fabs(pair.second->GetPos().GetX() - pos.x) < rad &&
            std::fabs(pair.second->GetPos().GetZ() - pos.z) < rad) {
            return true;
        }
    }
    return false;
}

void Room::CheckPartObjects(Human* testa, Human* testb)
{
    #if 0
    for (auto& pair1 : human_hash_) {
        Human* a = pair1.second;
        for (auto& pair2 : human_hash_) {
            Human* b = pair2.second;
            if (testa && testb) {
                if ((a == testa && b == testb) ||
                    (a == testb && b == testa)) {
                }
            }
            if (a->InPartObjects(b)) {
                if (!b->InPartObjects(a)) {
                    A8_ABORT();
                }
            }
        }
    }

    for (auto& pair1 : human_hash_) {
        Human* huma = pair1.second;
        huma->TraverseAllLayerHumanList
            (
             [&huma, testa, testb] (Human* humb, bool& stop)
             {
                if (testa && testb) {
                    if ((huma == testa && humb == testb) ||
                        (huma == testb && humb == testa)) {
                    }
                }
                if (a8::HasBitFlag(huma->status, CS_Disable) ||
                    a8::HasBitFlag(humb->status, CS_Disable) ||
                    huma->GetPartObjectsCount() <= 0
                    ) {
                    return;
                }
                if (!huma->InPartObjects(humb)) {
                    A8_ABORT();
                }
             });
    }
    #endif
}

bool Room::RuningInTimer()
{
    return xtimer.IsRunning();
}

#ifdef DEBUG
void Room::InitDebugInfo()
{
}

void Room::UnInitDebugInfo()
{
}
#endif

void Room::AddPlayerPostProc(Player* hum)
{
}

void Room::AddToEntityHash(Entity* entity)
{
    uniid_hash_[entity->GetUniId()] = entity;
}

void Room::AddToHumanHash(Human* hum)
{
    human_hash_[hum->GetUniId()] = hum;
}

void Room::AddToAliveHumanHash(Human* hum)
{
    alive_human_hash_[hum->GetUniId()] = hum;
    if (hum->IsPlayer()) {
        alive_player_hash_[hum->GetUniId()] = hum;
    }
}

void Room::AddToMoveableHash(MoveableEntity* entity)
{
    moveable_hash_[entity->GetUniId()] = entity;
}

void Room::AddToAccountHash(Player* hum)
{
    accountid_hash_[hum->account_id] = hum;
}

void Room::AddToLaterAddHash(RoomEntity* entity)
{
    later_add_hash_[entity->GetUniId()] = entity;
}

void Room::AddToRemovedRobotHash(Human* hum)
{
    removed_robot_hash_[hum->GetUniId()] = hum;
}

void Room::RemoveFromEntityHash(Entity* entity)
{
    uniid_hash_.erase(entity->GetUniId());
}

void Room::RemoveFromMoveableHash(MoveableEntity* entity)
{
    moveable_hash_.erase(entity->GetUniId());
}

void Room::RemoveFromHuamnHash(Human* hum)
{
    human_hash_.erase(hum->GetUniId());
}

void Room::RemoveFromAliveHumanHash(Human* hum)
{
    alive_human_hash_.erase(hum->GetUniId());
    if (hum->IsPlayer()) {
        alive_player_hash_.erase(hum->GetUniId());
    }
}

void Room::RemoveFromLaterAddHash(RoomEntity* entity)
{
    later_add_hash_.erase(entity->GetUniId());
}

void Room::CombineTeamBornPoint()
{
    for (auto& pair : team_hash_) {
        pair.second->CombineBornPoint();
    }
}

void Room::ForceSetBornPoint(Human* hum, std::shared_ptr<BornPoint> born_point)
{
    if (born_point && hum->born_point != born_point) {
        if (hum->born_point) {
            DecBornPointHumanNum(hum->born_point, hum);
        }
        hum->born_point = born_point;
        if (hum->born_point) {
            IncBornPointHumanNum(hum->born_point, hum);
            hum->SetPos(hum->born_point->RandPoint(this));
        }
        hum->FindLocation();
        hum->RefreshView();
        grid_service->MoveCreature(hum);
    }
}

std::shared_ptr<BornPoint> Room::ForceTakeBornPoint(Human* hum, std::shared_ptr<BornPoint> reserve_born_point)
{
    if (!reserve_born_point) {
        A8_ABORT();
    }
    if (hum->born_point == reserve_born_point) {
        A8_ABORT();
    }
    if (!hum->born_point) {
        A8_ABORT();
    }
    std::shared_ptr<BornPoint> pre_point = hum->born_point;
    if (pre_point) {
        DecBornPointHumanNum(pre_point, hum);
    }
    for (auto& pair : human_hash_) {
        if (pair.second->born_point == reserve_born_point) {
            ForceSetBornPoint(pair.second, pre_point);
        }
    }
    return reserve_born_point;
}

void Room::NewBieRoomStart()
{
}

Human* Room::GetOneCanEnableAndroid()
{
    std::vector<Human*> humans;
    GetCanEnableAndroids(humans, 1);
    return !humans.empty() ? humans[0] : nullptr;
}

void Room::GetCanEnableAndroids(std::vector<Human*>& humans, size_t num)
{
    for (auto& pair : human_hash_) {
        if (pair.second->IsAndroid() &&
            !pair.second->real_dead &&
            a8::HasBitFlag(pair.second->status, CS_Disable)
            ) {
            if (humans.size() >= num) {
                break;
            }
            humans.push_back(pair.second);
        }
    }
}

Player* Room::GetOneAlivePlayer()
{
    std::vector<Player*> humans;
    GetAlivePlayers(humans, 1);
    return !humans.empty() ? humans[0] : nullptr;
}

void Room::GetAlivePlayers(std::vector<Player*>& humans, size_t num)
{
    for (auto& pair : accountid_hash_) {
        if (!pair.second->real_dead) {
            if (humans.size() >= num) {
                break;
            }
            humans.push_back(pair.second);
        }
    }
}

int Room::GetCanShuaNum(int shua_num)
{
    int real_shua_num = shua_num;
    return real_shua_num;
}

bool Room::CanAddToScene(Human* hum)
{
    return true;
}

void Room::SyncFrameData()
{
#ifdef DEBUG
    long long begin_tick = a8::XGetTickCount();
    long long end_tick = a8::XGetTickCount();
#endif
    for (auto& pair : human_hash_) {
        pair.second->SendUpdateMsg();
#ifdef DEBUG
        end_tick = a8::XGetTickCount();
        #if 0
        if (a8::XGetTickCount() - begin_tick > 100) {
            A8_ABORT();
        }
        #endif
        begin_tick = a8::XGetTickCount();
#endif
    }
#ifdef DEBUG
    end_tick = a8::XGetTickCount();
    if (a8::XGetTickCount() - begin_tick > 1000) {
        A8_ABORT();
    }
    begin_tick = a8::XGetTickCount();
#endif
    frame_event.Clear();
}

void Room::RemoveRescue(Human* hum)
{
    for (auto& pair : human_hash_) {
        if (pair.second != hum && pair.second->GetActionType() == AT_Relive &&
            pair.second->GetActionTargetId() == hum->GetUniId()) {
            pair.second->CancelAction();
        }
    }
}

void Room::NotifyCountdown(const std::string& msg, int time)
{
    for (auto& pair : accountid_hash_) {
        pair.second->SendShowCountdown(msg, time);
    }
}

void Room::NotifySysPiao(const std::string& msg, int color, int duration)
{
    for (auto& pair : accountid_hash_) {
        pair.second->SendSysPiaoMsg(msg, color, duration);
    }
}

int Room::GetOnlinePlayerNum()
{
    int num = 0;
    for (auto& pair : accountid_hash_) {
        if (pair.second->socket_handle != 0) {
            ++num;
        }
    }
    return num;
}

size_t Room::GetRoomMaxPlayerNum()
{
    if (pve_instance) {
        return pve_human_num;
    }
    if (IsNewBieRoom()) {
        return 1;
    }
    if (custom_battle_) {
        return custom_battle_->GetMemberNum();
    }
    return map_meta_->player();
}

void Room::InitAndroidAI()
{
}

void Room::ForwardGasRing(int n)
{
    while (n > 0) {
        if (gas_data_.new_area_meta->IsLastGas()) {
            gas_data_.is_last_gas = true;
            return;
        }
        int pre_area_id = gas_data_.new_area_meta->id();
        gas_data_.SetGasMode(GasWaiting);
        glm::vec2 pre_pos = gas_data_.pos_new;
        gas_data_.old_area_meta = mt::SafeArea::GetById(pre_area_id);
        gas_data_.new_area_meta = mt::SafeArea::GetById(pre_area_id + 1);
        gas_data_.gas_progress = gas_data_.old_area_meta->rad();
        gas_data_.gas_start_frameno = GetFrameNo();
        gas_data_.pos_old = pre_pos;
        gas_data_.pos_old_bk = gas_data_.pos_old;
        {
            bool gen_ok = GenSmallCircle();
            if (!gen_ok) {
                A8_ABORT();
            }
        }
        gas_data_.rad_old = gas_data_.old_area_meta->rad();
        gas_data_.rad_new = gas_data_.new_area_meta->rad();
        gas_data_.gas_count++;
        --n;
    }
    if (gas_data_.old_area_meta) {
        auto boss_tuple = gas_data_.old_area_meta->GetBoss();
        if (boss_tuple && !IsGameOver()) {
            const mt::Hero* hero_meta = mt::Hero::GetById(std::get<2>(*boss_tuple));
            if (hero_meta) {
                int boss_uniid = AllocUniid();
                NotifyNewsTicker
                    (2,
                     {
                         a8::XValue(boss_uniid).GetString(),
                             a8::XValue(hero_meta->id()).GetString(),
                             a8::XValue(std::get<0>(*boss_tuple).x).GetString(),
                             a8::XValue(std::get<0>(*boss_tuple).y).GetString(),
                             a8::XValue(std::get<0>(*boss_tuple).z).GetString(),
                             a8::XValue(std::get<1>(*boss_tuple)).GetString(),
                         });
                xtimer.SetTimeoutEx
                    (SERVER_FRAME_RATE * std::get<1>(*boss_tuple),
                     [this, boss_tuple, hero_meta, boss_uniid] (int event, const a8::Args* args)
                     {
                         if (a8::TIMER_EXEC_EVENT == event) {
                             Hero* hero = CreateHero
                                 (nullptr,
                                  hero_meta,
                                  std::get<0>(*boss_tuple),
                                  GlmHelper::UP,
                                  666,
                                  boss_uniid);
                             if (hero) {
                                 batch_sync_->AddGlobalObject(hero);
                                 hero->GetTrigger()->AddListener
                                     (
                                      kDieEvent,
                                      [this, boss_uniid] (const a8::Args& args) mutable
                                      {
                                          if (!IsGameOver()) {
                                              batch_sync_->RemoveGlobalObject(boss_uniid);
                                              NotifyKillList({boss_uniid});
                                          }
                                      });
                             }
                         }
                     },
                     &xtimer_attacher_);
            }
        }
    }
}

void Room::GetPartObjectWatchList(Entity* entity, std::vector<Human*>& watch_list)
{
    TraverseHumanList
        (
         [&watch_list, entity] (Human* hum) -> bool
         {
             if (hum->InPartObjects(entity)) {
                 watch_list.push_back(hum);
             }
             return true;
         });
}

void Room::SetInfiniteBulletMode()
{
    infinite_bullet_mode_ = true;
    for (auto& pair : accountid_hash_) {
        pair.second->SetInfiniteBulletMode();
    }
}

void Room::AddToPostBattleAutoFreeList(a8::XTimerWp timer)
{
    if (BattleStarted()) {
        A8_ABORT();
    }
    if (timer.expired()) {
        A8_ABORT();
    }
    // 999
    #if 1
    #else
    if (post_battle_auto_free_list_.find(timer) != post_battle_auto_free_list_.end()) {
        A8_ABORT();
    }
    #endif
    post_battle_auto_free_list_.push_back(timer);
}

void Room::ClearPostBattleAutoFreeList()
{
    size_t i = 0;
    while (i < post_battle_auto_free_list_.size()) {
        auto timer = post_battle_auto_free_list_[i];
        if (!timer.expired()) {
            xtimer.Delete(timer);
        }
        ++i;
    }
    post_battle_auto_free_list_.clear();
}

void Room::OnBattleStart()
{
    battle_starting_ = true;
    std::vector<EntityWeakPtr> entitys;
    entitys.reserve(uniid_hash_.size());
    for (auto& pair : uniid_hash_) {
        entitys.push_back(pair.second->GetEntityWeakPtrRef());
    }
    for (auto& e : entitys) {
        if (e.Get()) {
            e.Get()->OnBattleStart(this);
        }
    }
    battle_starting_ = false;
    if (IsMiniMap() || IsNewerMap()) {
        xtimer.SetIntervalEx
            (
             SERVER_FRAME_RATE * 10,
             [this] (int event, const a8::Args* args)
             {
                 if (a8::TIMER_EXEC_EVENT == event) {
                     int keep_num = 0;
                     if (IsMiniMap()) {
                         keep_num = 2;
                         Player* hum = GetOneAlivePlayer();
                         if (hum && hum->GetBattleContext()->GetHeroLevel() > 5) {
                             keep_num = 3;
                         }
                     } else if (IsNewerMap()) {
                         keep_num = 1;
                     }
                     std::set<Human*> humans;
                     for (auto& pair : alive_human_hash_) {
                         if (keep_num <= 0) {
                             break;
                         }
                         if (a8::HasBitFlag(pair.second->status, CS_CrazeMode)) {
                             --keep_num;
                         } else {
                             humans.insert(pair.second);
                         }
                     }
                     std::map<int, Human*> players = alive_player_hash_;
                     while (keep_num > 0) {
                         Human* target = nullptr;
                         if (!players.empty()) {
                             for (auto& pair : players) {
                                 target = pair.second;
                                 break;
                             }
                         }
                         Human* craze_hum = nullptr;
                         if (target) {
                             for (auto& itr : humans) {
                                 if (itr->IsPlayer() || itr->GetTeam()->HasPlayer()) {
                                     continue;
                                 }
                                 if (!craze_hum) {
                                     craze_hum = itr;
                                 } else {
                                     if (craze_hum->GetPos().ManhattanDistance2D(target->GetPos()) >
                                         itr->GetPos().ManhattanDistance2D(target->GetPos())) {
                                         craze_hum = itr;
                                     }
                                 }
                             }
                             players.erase(target->GetUniId());
                         } else {
                             for (auto& itr : humans) {
                                 craze_hum = itr;
                                 break;
                             }
                         }
                         if (craze_hum) {
                             a8::SetBitFlag(craze_hum->status, CS_CrazeMode);
                             humans.erase(craze_hum);
#ifdef DEBUG
                             a8::XPrintf("craze huamn :%s\n", {craze_hum->name});
#endif
                         }
                         --keep_num;
                     }
                 }
             },
             &xtimer_attacher_);
    }
}

void Room::AddTeam(class MatchTeam* team)
{
    if (team->GetCurrMembers().size() <= 0) {
        abort();
    }
    if (team->GetCurrMembers().size() != team->GetSlotNum()) {
#if 1
        f8::UdpLog::Instance()->Warning
            ("addTeam currMembers.size:%d slot_num:%d",
             {
                 team->GetCurrMembers().size(),
                     team->GetSlotNum()
             });

#else
        return;
#endif
    }
    std::shared_ptr<Team> new_team = NewTeam();
    {
        new_team->SetInitTeamMemberNum(team->GetSlotNum());
        #if 0
        new_team->SetAutoFill(true);
        #endif
    }
    std::shared_ptr<BornPoint> init_born_point = nullptr;
    for (auto& member : team->GetCurrMembers()) {
        cs::CMJoin& msg = *member->msg;
        if (member->is_robot) {
            const mt::Robot* robot_meta = member->robot_meta;
            Android* hum = EntityFactory::Instance()->MakeAndroid(AllocUniid());
            #if 1
            a8::SetBitFlag(hum->status, CS_ForceTeam);
            #endif
            hum->name = robot_meta->name();
            hum->meta = mt::Hero::GetById(robot_meta->hero_id());
            hum->robot_meta = robot_meta;
            if (!init_born_point) {
                init_born_point = AllocBornPoint(hum);
            }
            hum->born_point = init_born_point;
            Global::Instance()->verify_set_pos = 1;
            if (!hum->born_point) {
                abort();
            } else {
                hum->SetPos(hum->born_point->RandPoint(this));
            }
            Global::Instance()->verify_set_pos = 0;
            glm::vec3 attack_dir = hum->GetPos().ToGlmVec3();
            GlmHelper::Normalize(attack_dir);
            GlmHelper::RotateY(attack_dir, a8::RandAngle());
            hum->SetAttackDir(attack_dir);
            hum->SetMoveDir(attack_dir);
            hum->room = this;
            hum->Initialize();
            AddToEntityHash(hum);
            AddToHumanHash(hum);
            #if 0
            MatchTeam(hum);
            #endif
            ++alive_count_;
            alive_count_chged_frameno_ = GetFrameNo();
            ++PerfMonitor::Instance()->alive_count;
            refreshed_robot_set_.insert(robot_meta->id());
            {
                AddToAliveHumanHash(hum);
                AddToMoveableHash(hum);
                grid_service->AddCreature(hum);
                hum->FindLocation();
                hum->RefreshView();
            }
            new_team->AddMember(hum);
#ifdef DEBUG1
            if (!f8::IsTestEnv()) {
                hum->SetHP(50);
            }
#endif
        } else {
            Player* hum = NewPlayer();
            hum->proto_version = msg.proto_version();
            hum->hero_uniid = a8::XValue(msg.hero_uniid());
            hum->battle_uuid = member->battle_uuid;
            hum->is_valid_battle = member->is_valid_battle;
            hum->payload = member->payload;
            #if 1
            a8::SetBitFlag(hum->status, CS_ForceTeam);
            #if 0
            member->msg.mutable_baseskin()->Clear();
            member->msg.add_baseskin(1);
            #endif
            #endif
            #if 0
            hum->ProcPreSettlementInfo(member->msg->pre_settlement_info());
            #endif
            PlayerMgr::Instance()->
                CreatePlayerByCMJoin(hum,
                                     0,//hdr.ip_saddr,
                                     member->socket_handle,
                                     *member->msg
                                     );
            hum->meta = mt::Hero::GetById(msg.hero_id());
            if (!hum->meta) {
                hum->meta = mt::Param::s().human_meta;
            }
            hum->room = this;
            hum->SetBattleContext(member->battle_context);
            AddPlayer(hum, init_born_point, true);
            hum->GetBattleContext()->Init(hum);
            hum->ProcSkillList();
            hum->SetHP(hum->GetBattleContext()->GetMaxHP());
            hum->SetMaxHP(hum->GetHP());
            PlayerMgr::Instance()->IncAccountNum(msg.account_id());
            new_team->AddMember(hum);
        }
    }
    NotifyUiUpdate();
}

bool Room::CanAddObstacle(const glm::vec3& pos, int obstacle_id)
{
    const mt::MapThing* thing_meta = mt::MapThing::GetById(obstacle_id);
    if (!thing_meta) {
        return false;
    }
    int rad = std::max(thing_meta->width(), thing_meta->height());
    return map_service->CanAdd(pos, rad);
}

bool Room::IsDestoryRoom()
{
    return GetMapMeta()->map_id() == 1002;
}

bool Room::IsSurvivalRoom()
{
    return GetMapMeta()->map_id() == 1003;
}

void Room::OnEnterNewWave(int wave)
{
    if (IsDestoryRoom() && wave > 0 && wave < pve_mode_meta->_area.size()) {
        gas_data_.SetGasMode(GasWaiting);
        gas_data_.gas_start_frameno = GetFrameNo();
        gas_data_.old_area_meta = pve_mode_meta->_area[wave];
        gas_data_.new_area_meta = pve_mode_meta->_area[wave];
        gas_data_.gas_start_frameno = GetFrameNo();
        gas_data_.pos_new = glm::vec2(gas_data_.new_area_meta->x1(),
                                      gas_data_.new_area_meta->y1());
        gas_data_.pos_old = gas_data_.pos_new;
        gas_data_.pos_old_bk = gas_data_.pos_old;
        gas_data_.rad_old = gas_data_.old_area_meta->rad();
        gas_data_.rad_new = gas_data_.new_area_meta->rad();
        gas_data_.gas_progress = 0;
    }
    NotifyUiUpdate();
#ifdef DEBUG
        a8::XPrintf("OnEnternerWave:%d\n", {wave});
#endif
}

bool Room::IsAllRealDead()
{
    bool is_all_dead = true;
    TraverseHumanList
        (
         [&is_all_dead] (Human* hum)
         {
             if (!hum->real_dead) {
                 is_all_dead = false;
                 return false;
             }
             return true;
         }
         );
    return is_all_dead;
}

void Room::AddTask(int task_uniid, std::shared_ptr<a8::CommonCbProcEx> cb)
{
    if (task_hash_.find(task_uniid) != task_hash_.end()) {
        abort();
    }
    task_hash_[task_uniid] = cb;
}

void Room::RemoveTask(int task_uniid)
{
    task_hash_.erase(task_uniid);
}

void Room::AutoJump()
{
    int auto_jump_min_num = mt::Param::GetIntParam("auto_jump_min_num");
    int auto_jump_max_num = mt::Param::GetIntParam("auto_jump_max_num");
    int jump_num = a8::RandEx(auto_jump_min_num, auto_jump_max_num);
    if (GlmHelper::Norm(last_player_jump_pos - plane.curr_pos)< 64 * 8) {
        jump_num = 1 + rand() % 2;
    }
    for (int i = 0; i < jump_num; ++i) {
        TraverseHumanList
            (
             [] (Human* hum) -> bool
             {
                 if (hum->HasBuffEffect(kBET_Fly) &&
                     hum->GetEntitySubType() != EST_Player) {
                     hum->DoJump();
                     return false;
                 }
                 return true;
             });
    }
}

int Room::GetPolyExtFlag(int poly_idx)
{
    if (poly_idx >=0 && poly_idx < poly_ext_datas_.size()) {
        return poly_ext_datas_.at(poly_idx);
    }
    return 0;
}

void Room::CreateWorldObjects()
{
    if (IsNewBieRoom()) {
        int uniid = AllocUniid();
        std::shared_ptr<BornPoint> born_point = std::make_shared<BornPoint>();
        std::shared_ptr<mt::WorldObject> wo_meta = std::make_shared<mt::WorldObject>();
        wo_meta->pos = mt::Param::s().newbie_room_born_point;
        born_point->wo_meta = wo_meta;
        born_point_hash_[uniid] = born_point;
        return;
    }
    if (IsPveRoom()) {
        int uniid = AllocUniid();
        std::shared_ptr<BornPoint> born_point = std::make_shared<BornPoint>();
        std::shared_ptr<mt::WorldObject> wo_meta = std::make_shared<mt::WorldObject>();
        wo_meta->pos = pve_instance->_spawn_point;
        born_point->wo_meta = wo_meta;
        born_point_hash_[uniid] = born_point;
        return;
    }
    for (auto& itr : GetMapMeta()->_world_objects) {
        std::shared_ptr<mt::WorldObject> obj = itr;
        switch ((WorldObjectType_e)obj->object_type) {
        case WorldObjectType_e::kLootType:
            {
                CreateLoot
                    (
                     obj->object_id,
                     obj->pos,
                     obj->pos,
                     1,
                     1,
                     true);
            }
            break;
        case WorldObjectType_e::kBoxType:
            {
                CreateObstacle
                    (
                     obj->object_id,
                     obj->pos.x,
                     obj->pos.y,
                     obj->pos.z,
                     std::make_shared<a8::Args>(std::vector<std::any>({obj}))
                     );
            }
            break;
        case WorldObjectType_e::kCarType:
            {
                const mt::Equip* equip_meta = mt::Equip::GetById(obj->object_id);
                if (equip_meta) {
                    int car_uniid = AllocUniid();
                    glm::vec3 pos = obj->pos - obj->size / 2.0f;
                    Car* c = CreateCar(nullptr,
                                       car_uniid,
                                       equip_meta,
                                       pos,
                                       0,
                                       nullptr);
                    if (c) {
                        CarObject car;
                        car.car_id = equip_meta->id();
                        car.pos = c->GetPos().ToGlmVec3();
                        car_hash_[c->GetUniId()] = car;
                    }
                }
            }
            break;
        case WorldObjectType_e::kBornPointType:
            {
                int uniid = AllocUniid();
                std::shared_ptr<BornPoint> born_point = std::make_shared<BornPoint>();
                born_point->wo_meta = obj;
                born_point_hash_[uniid] = born_point;
            }
            break;
        default:
            {
            }
            break;
        }
    }
    for (auto& itr : GetMapMeta()->_group_world_objects) {
        int ikey = itr.first;
        auto objects = itr.second;
        std::shuffle(objects.begin(), objects.end(), std::default_random_engine(a8::XGetTickCount()));
        const mt::MapThingGroup* group_meta = mt::MapThingGroup::GetById(ikey);
        if (group_meta) {
            int i = 0;
            for (auto& tuple : *group_meta->GetItems()) {
                if (i < objects.size()) {
                    switch (std::get<0>(tuple)) {
                    case 1:
                        {
                            const mt::Equip* equip_meta = mt::Equip::GetById(std::get<1>(tuple));
                            if (equip_meta) {
                                CreateLoot
                                    (
                                     std::get<1>(tuple),
                                     objects[i]->pos,
                                     objects[i]->pos,
                                     1,
                                     1
                                     );
                            } else {
                                const mt::MapThing* thing_meta = mt::MapThing::GetById(std::get<1>(tuple));
                                if (thing_meta) {
                                    if (thing_meta->thing_type() == 16) {
                                        for (int drop_id : thing_meta->RandDrop()) {
                                            ScatterDrop(objects[i]->pos, drop_id);
                                        }
                                    } else {
                                        CreateObstacle
                                            (
                                             std::get<1>(tuple),
                                             objects[i]->pos.x,
                                             objects[i]->pos.y,
                                             objects[i]->pos.z
                                             );
                                    }
                                }
                            }
                        }
                        break;
                    case 2:
                        {
                            #ifdef DEBUG1
                            #else
                            const mt::Hero* hero_meta = mt::Hero::GetById(std::get<1>(tuple));
                            if (hero_meta) {
                                Position hero_pos;
                                hero_pos.FromGlmVec3(objects[i]->pos);

                                int team_id = 666;
                                Creature* master = nullptr;

                                glm::vec3 dir = GlmHelper::UP;

                                Hero* hero = CreateHero(master,
                                                        hero_meta,
                                                        hero_pos.ToGlmVec3(),
                                                        dir,
                                                        team_id);
                            }
                            #endif
                        }
                        break;
                    default:
                        {
                        }
                        break;
                    } //end switch
                }
                ++i;
            }
        }
    }
    if (born_point_hash_.size() < 10) {
        #if 1
        while (born_point_hash_.size() < 10) {
            int uniid = AllocUniid();
            std::shared_ptr<BornPoint> born_point = std::make_shared<BornPoint>();
            std::shared_ptr<mt::WorldObject> wo_meta = std::make_shared<mt::WorldObject>();
            wo_meta->pos.x = (60 + rand() % ((int)GetMapMeta()->map_width() - 100));
            wo_meta->pos.y = 6.250846862793;
            wo_meta->pos.z = (60 + rand() % ((int)GetMapMeta()->map_height() - 100));
            born_point->wo_meta = wo_meta;
            born_point_hash_[uniid] = born_point;
        }
        #else
        abort();
        #endif
    }
}

Team* Room::GetTeam(int team_id)
{
    auto itr = team_hash_.find(team_id);
    return itr != team_hash_.end() ? itr->second.get() : nullptr;
}

Creature* Room::GetCreatureByUniId(int uniid)
{
    Entity* e = GetEntityByUniId(uniid);
    return e && e->IsCreature(this) ? (Creature*)e : nullptr;
}

bool Room::RandomSafeAreaPoint(glm::vec3& point)
{
    glm::vec3 dir = GlmHelper::UP;
    GlmHelper::RotateY(dir, a8::RandAngle());
    float rnd_rad = rand() % std::max(GetGasData().new_area_meta->rad(), 100);
    glm::vec2 center = GetGasData().pos_new + glm::vec2(dir.x, dir.z) * (100.f + rnd_rad);
    point = glm::vec3(center.x, 0.0f, center.y);
    return true;
}

bool Room::FindWalkablePointWithOutHouse(const glm::vec3& center, int try_count, int step, glm::vec3& out_point)
{
    glm::vec3 ref_point = center;
    map_instance->PtInHouse(center, ref_point);
    map_instance->Scale(ref_point);
    glm::vec3 point;
    bool ok = false;
    for (int i = 0; i < try_count; ++i) {
        ok = map_instance->FindNearestPoint(ref_point, 1.0f + step * i, point);
        if (ok) {
            break;
        }
    }
    if (!ok) {
        return false;
    }
    map_instance->UnScale(point);
    out_point = point;
    return true;
}

void Room::AddReportBullet(int bullet_uniid)
{
    report_bullet_hash[bullet_uniid] = GetFrameNo();
    xtimer.SetTimeoutEx
        (
         SERVER_FRAME_RATE * 10,
         [this, bullet_uniid] (int event, const a8::Args* args)
         {
             if (a8::TIMER_EXEC_EVENT == event) {
                 report_bullet_hash.erase(bullet_uniid);
                 Entity* entity = GetEntityByUniId(bullet_uniid);
                 if (entity && entity->GetEntityType() == ET_Bullet) {
                     Bullet* bullet = (Bullet*)entity;
                     bullet->ForceRemove();
                 }
             }
         },
         &xtimer_attacher_);
}

int Room::GetReportRoomMode()
{
    return IsPveRoom() ? 1 : 0;
}

int Room::GetPvpMatchMode()
{
    return GetRoomMode() == kPvpRankMode ? 1 : 0;
}

void Room::ForceOver()
{
    if (added_to_over_room) {
        return;
    }
    std::vector<Team*> alive_teams;
    TraverseTeams
        (
         [&alive_teams] (Team* team)
         {
             if (team->HasAliveMember()) {
                 alive_teams.push_back(team);
             }
             return true;
         });
    std::sort(alive_teams.begin(), alive_teams.end(),
              [] (Team* a, Team *b) -> bool
              {
                  if (a->HasPlayer()) {
                      return true;
                  }
                  return false;
              });
    for (Team* team : alive_teams) {
        team->TraverseMembers
            (
             [] (Human* member)
             {
                 if (!member->dead) {
                     member->BeKill(VP_Gas,
                                    TEXT("battle_server_killer_gas", "毒圈"),
                                    VW_Gas,
                                    VP_Gas,
                                    TEXT("battle_server_killer_gas", "毒圈"));
                 }
                 return true;
             });
    }
    force_over_ = true;
    GameLog::Instance()->ForceOver(this);
    xtimer.SetTimeoutEx
        (SERVER_FRAME_RATE * 5,
         [this] (int event, const a8::Args* args)
         {
             if (a8::TIMER_EXEC_EVENT == event) {
                 if (!game_over_) {
                     game_over_ = true;
                     game_over_frameno_ = GetFrameNo();
                     OnGameOver();
                 }
             }
         },
         &xtimer_attacher_);
}

bool Room::IsNewBieRoom()
{
    return is_newbie_room_;
}

bool Room::SupportSandTable()
{
    return PBUtils::SupportSandTable(creator_proto_version_);
}

void Room::OnAddHuman(Human* hum)
{
    if (IsSandTableRoom()) {
        hum->TryAddBuff(hum, kThroughWall_BUFFID);
    }
}

bool Room::IsSandTableRoom()
{
    return !IsPveRoom() &&
        !IsNewBieRoom() &&
        SupportSandTable();
}

bool Room::HasRoomSwitch(int tag)
{
    return a8::HasBitFlag(room_switch_, tag);
}

void Room::OpenRoomSwitch(int tag)
{
    a8::SetBitFlag(room_switch_, tag);
}

void Room::CloseRoomSwitch(int tag)
{
    a8::UnSetBitFlag(room_switch_, tag);
}

void Room::GMFastForward()
{
    while (GetGasData().new_area_meta->GetSmallRingCount() >= 3) {
        ForwardGasRing(1);
    }
    TraversePlayerList
        (
         [this] (Player* hum)
         {
             glm::vec3 src_point = glm::vec3(GetGasData().new_area_meta->GetLastArea()->x1(),
                                         6.0f,
                                         GetGasData().new_area_meta->GetLastArea()->y1());
             {
                 glm::vec3 center = src_point;
                 map_instance->Scale(center);
                 glm::vec3 point;
                 bool ok = false;
                 for (int i = 0; i < 10; ++i) {
                     ok = map_instance->FindConnectableNearestPoint(center, 1.0f + 10 * i, point);
                     if (ok) {
                         break;
                     }
                 }
                 if (ok) {
                     map_instance->UnScale(point);
                     hum->GetMutablePos().FromGlmVec3(point);
                     grid_service->MoveCreature(hum);
                 }
             }
         });
    GetIncubator()->Clear(5);
    for (auto& pair : human_hash_) {
        if (pair.second->IsAndroid() &&
            !a8::HasBitFlag(pair.second->status, CS_Disable)) {
            Human* hum = pair.second;
            hum->BeKill(VP_Gas,
                        TEXT("battle_server_killer_gas", "毒圈"),
                        VW_Gas,
                        VP_Gas,
                        TEXT("battle_server_killer_gas", "毒圈"));
        }
    }
    xtimer.SetTimeoutEx
        (
         1,
         [this] (int event, const a8::Args* args)
         {
             if (a8::TIMER_EXEC_EVENT == event) {
                 GetIncubator()->ShowHand();
             }
         },
         &xtimer_attacher_);
}

void Room::LockRoom()
{
    lock_room_ = true;
    lock_room_frameno_ = GetFrameNo();
}

bool Room::BulletCanReach(const glm::vec3& start, const glm::vec3& end)
{
    glm::vec3 a_start = start;
    glm::vec3 a_end = end;
    glm::vec3 hit_point;
    bool hit_result = false;

    map_instance->Scale(a_start);
    map_instance->Scale(a_end);
    if (map_instance->Raycast(a_start, a_end, hit_point, hit_result)) {
        if (!hit_result) {
            return true;
        }
    }
    return false;
}

bool Room::MoveCanReach(const glm::vec3& start, const glm::vec3& end)
{
    return BulletCanReach(start, end);
}

bool Room::IsPvpMasterRankModeRoom()
{
    return rank_match_conf_ && rank_match_conf_->room_type == RoomType_MasterRank;
}

int Room::GetMaxTeamNum()
{
    return MAX_TEAM_NUM;
}

void Room::FillCustomBattleHuman()
{
    Player* target = nullptr;
    for (auto& pair : accountid_hash_) {
        target = pair.second;
        break;
    }
    if (!target) {
        abort();
    }
    custom_battle_->TraverseMemberList
        (
         [this] (CustomMember* member) -> bool
         {
             if (!member->IsJoined()) {
                 int socket_handle = 0;
                 int ip_saddr = 0;
                 member->GetNetData()->join_msg = std::make_shared<cs::CMJoin>();
                 member->GetNetData()->join_msg->set_account_id(member->GetAccountId());

                 Player* hum = NewPlayer();
                 hum->proto_version = member->GetNetData()->join_msg->proto_version();
#if 0
                 hum->hero_uniid = a8::XValue(msg.hero_uniid());
#endif
                 hum->battle_uuid = member->GetNetData()->battle_uuid;
                 hum->is_valid_battle = member->GetNetData()->is_valid_battle;
                 hum->payload = member->GetNetData()->payload;
                 member->GetNetData()->join_msg->set_session_id(member->GetSessionId());
                 member->GetNetData()->join_msg->set_team_uuid
                     (member->GetTeam()->GetTeamUuid());
                 PlayerMgr::Instance()->
                     CreatePlayerByCMJoin(hum,
                                          ip_saddr,
                                          socket_handle,
                                          *member->GetNetData()->join_msg
                                          );
                 hum->meta = mt::Hero::GetById(member->GetNetData()->join_msg->hero_id());
                 if (!hum->meta) {
                     hum->meta = mt::Param::s().human_meta;
                 }
                 hum->room = custom_battle_->GetRoom();
                 hum->SetBattleContext(member->GetNetData());
                 hum->GetBattleContext()->Init(hum);
                 {
                     long long hero_uniid = 0;
                     int hero_lv = 1;
                     int quality = 1;
                     hum->GetBattleContext()->GetHeroLvQuality(hero_uniid, hero_lv, quality);
                     hum->hero_uniid = hero_uniid;
                 }
                 custom_battle_->GetRoom()->AddPlayer(hum);
                 hum->ProcSkillList();
                 hum->SetHP(hum->GetBattleContext()->GetMaxHP());
                 hum->SetMaxHP(hum->GetHP());
                 member->Join(hum);
                 PlayerMgr::Instance()->IncAccountNum(hum->account_id);
             }
             return true;
         });
}


bool Room::IsMiniMap()
{
    return map_instance->map_id == 2002;
}

bool Room::IsNewerMap()
{
    return map_instance->map_id == 2003;
}

int Room::GetMobaLeftTime()
{
    if (moba_over_timer.expired()) {
        return 0;
    }
    long long remain_time = xtimer.GetRemainTime(moba_over_timer);
    return (remain_time * FRAME_RATE_MS);
}
